from __future__ import annotations

import logging
import re
from functools import reduce

from libtbx.phil import parse
from scitbx import sparse
from scitbx.array_family import flex

from dials.algorithms.refinement import DialsRefineConfigError

logger = logging.getLogger(__name__)

# PHIL options for constraints
phil_str = """
constraints
  .help = "Parameter equal shift constraints to use in refinement."
  .expert_level = 2
  .multiple = True
{
  id = None
    .help = "Select only the specified experiments when looking up which"
            "parameterisations to apply the constraint to. If an identified"
            "parameterisation affects multiple experiments then the index"
            "of any one of those experiments suffices to identify that"
            "parameterisation. If None (the default) then constraints will be"
            "applied to all parameterisations of this type."
    .type = ints(value_min=0)

  parameter = None
    .type = str
    .help = "Identify which parameter of each parameterisation to constrain by"
            "a (partial) parameter name to match. Model name prefixes such as"
            "'Detector1' will be ignored as parameterisations are already"
            "identified by experiment id"
}
"""

phil_scope = parse(phil_str)


class EqualShiftConstraint:
    """A single constraint between parameters of the same type in different
    parameterisations"""

    def __init__(self, indices, parameter_vector):

        self.indices = indices
        parameter_vector = flex.double(parameter_vector)
        self.constrained_value = flex.mean(parameter_vector.select(indices))
        self._shifts = parameter_vector.select(indices) - self.constrained_value

    def set_constrained_value(self, val):
        self.constrained_value = val

    def get_expanded_values(self):
        return self.constrained_value + self._shifts


class ConstraintManager:
    def __init__(self, constraints, n_full_params):

        self._constraints = [e for e in constraints if e is not None]

        # constraints should be a list of EqualShiftConstraint objects
        assert len(self._constraints) > 0

        self._n_full_params = n_full_params
        full_idx = flex.size_t_range(n_full_params)
        self._constrained_gps = [c.indices for c in self._constraints]
        self._constrained_idx = flex.size_t(
            [i for c in self._constrained_gps for i in c]
        )
        keep = flex.bool(self._n_full_params, True)
        keep.set_selected(self._constrained_idx, False)
        self._unconstrained_idx = full_idx.select(keep)
        self._n_unconstrained_params = len(self._unconstrained_idx)

    def get_constrained_parameter_indices(self):
        return [c.indices for c in self._constraints]

    def constrain_parameters(self, x):

        assert len(x) == self._n_full_params

        constrained_vals = flex.double([c.constrained_value for c in self._constraints])

        # select unconstrained parameters only
        unconstrained_x = x.select(self._unconstrained_idx)

        constrained_x = flex.double.concatenate(unconstrained_x, constrained_vals)

        return constrained_x

    def expand_parameters(self, constrained_x):

        unconstrained_part = constrained_x[0 : self._n_unconstrained_params]
        constrained_part = constrained_x[self._n_unconstrained_params :]

        # update constrained parameter values
        for v, c in zip(constrained_part, self._constraints):
            c.set_constrained_value(v)

        expanded = flex.double(
            [v for c in self._constraints for v in c.get_expanded_values()]
        )

        full_x = flex.double(self._n_full_params)
        full_x.set_selected(self._unconstrained_idx, unconstrained_part)
        full_x.set_selected(self._constrained_idx, expanded)

        return full_x

    def constrain_jacobian(self, jacobian):

        # set up result matrix
        nrow = jacobian.all()[0]
        ncol = self._n_unconstrained_params + len(self._constraints)
        constrained_jacobian = flex.double(flex.grid(nrow, ncol))

        # create constrained columns
        constr_block = flex.double(flex.grid(nrow, len(self._constraints)))
        for i, gp in enumerate(self._constrained_gps):
            cols = [jacobian.matrix_copy_column(j) for j in gp]
            sum_col = reduce(lambda x, y: x + y, cols)
            constr_block.matrix_paste_column_in_place(sum_col, i)

        # copy unconstrained columns into the result
        for i, j in enumerate(self._unconstrained_idx):
            col = jacobian.matrix_copy_column(j)
            constrained_jacobian.matrix_paste_column_in_place(col, i)

        # copy the constrained block into the result
        constrained_jacobian.matrix_paste_block_in_place(
            constr_block, 0, self._n_unconstrained_params
        )

        return constrained_jacobian

    def constrain_gradient_vector(self, grad):

        # extract unconstrained gradients into the result
        result = []
        result = list(flex.double(grad).select(self._unconstrained_idx))

        # append constrained gradients
        for i, gp in enumerate(self._constrained_gps):
            vals = [grad[j] for j in gp]
            result.append(sum(vals))

        return result


class SparseConstraintManager(ConstraintManager):
    def constrain_jacobian(self, jacobian):
        """sparse matrix version of constrain_jacobian"""

        # select unconstrained columns only
        unconstr_block = jacobian.select_columns(self._unconstrained_idx)

        # create constrained columns
        constr_block = sparse.matrix(jacobian.n_rows, len(self._constraints))

        for i, (gp, c) in enumerate(zip(self._constrained_gps, constr_block.cols())):
            # this copies, so c is no longer the matrix column but a new vector
            for j in gp:
                c += jacobian.col(j)
            # so assign back into the matrix directly
            constr_block[:, i] = c

        # construct the constrained Jacobian
        constrained_jacobian = sparse.matrix(
            jacobian.n_rows, unconstr_block.n_cols + constr_block.n_cols
        )
        constrained_jacobian.assign_block(unconstr_block, 0, 0)
        constrained_jacobian.assign_block(constr_block, 0, unconstr_block.n_cols)

        return constrained_jacobian


class ConstraintManagerFactory:
    """Build equal shift constraints as requested in params and package into
    a constraints manager to be linked to the Refinery"""

    def __init__(self, refinement_phil, pred_param, sparse=False):

        self._params = refinement_phil
        self._pred_param = pred_param

        return

    def build_constraint(self, constraint_scope, parameterisation, model_type):
        """Create a constraint for a single parameter specified by
        constraint_scope"""

        if constraint_scope.id is None:
            # get one experiment id for each parameterisation to apply to all
            constraint_scope.id = [e.get_experiment_ids()[0] for e in parameterisation]

        # find which parameterisations are involved
        prefixes = []
        for i, p in enumerate(parameterisation):
            if hasattr(p, "num_samples"):
                raise DialsRefineConfigError(
                    "Constraints cannot be created between scan-varying "
                    "parameterisations."
                )
            for j in p.get_experiment_ids():
                if j in constraint_scope.id:
                    prefixes.append(model_type + f"{i + 1}")
                    break
        logger.debug(
            "\nSelection of experiments by id has identified the following "
            "models to link by constraints:\n  " + "\n  ".join(prefixes)
        )

        # ignore model name prefixes
        patt1 = re.compile("^" + model_type + "[0-9]+")
        pname = patt1.sub("", constraint_scope.parameter)

        # Use a regex to find the parameters to constrain from a list of all the
        # parameter names. There are multiple parts to this. The first part
        # identifies the relevant model type and parameterisation ordinal index,
        # accepting those that were chosen according to the supplied experiment
        # ids. The next part allows for additional text, like 'Group1' that may
        # be used by a multi-panel detector parameterisation. Then the parameter
        # name itself, like 'Dist'.
        patt2 = re.compile(
            "^(" + "|".join(prefixes) + r"){1}(?![0-9])(\w*" + pname + ")$"
        )
        indices = [j for j, s in enumerate(self._all_names) if patt2.match(s)]
        if len(indices) < 2:
            logger.warn("No pair of parameters to link by constraint")
            return None
        logger.debug(
            "The following parameters will be constrained "
            "to enforce equal shifts at each step of refinement:\n  "
            + "\n  ".join([self._all_names[k] for k in indices])
        )

        return EqualShiftConstraint(indices, self._all_vals)

    def __call__(self):

        # shorten options path
        options = self._params.refinement.parameterisation

        # shorten options paths further for individual parameterisation types
        detector_c = options.detector.constraints
        beam_c = options.beam.constraints
        orientation_c = options.crystal.orientation.constraints
        cell_c = options.crystal.unit_cell.constraints

        # full parameter names and values
        self._all_names = self._pred_param.get_param_names()
        self._all_vals = self._pred_param.get_param_vals()

        # quit early if there are no constraints to apply
        n_constraints = sum(len(e) for e in [detector_c, beam_c, orientation_c, cell_c])
        if n_constraints == 0:
            return None

        logger.debug("\nConfiguring constraints")

        # list of constraint objects to build up
        constraints = []

        detector_p = self._pred_param.get_detector_parameterisations()
        beam_p = self._pred_param.get_beam_parameterisations()
        orientation_p = self._pred_param.get_crystal_orientation_parameterisations()
        cell_p = self._pred_param.get_crystal_unit_cell_parameterisations()

        for constr in detector_c:
            constraints.append(self.build_constraint(constr, detector_p, "Detector"))
        for constr in beam_c:
            constraints.append(self.build_constraint(constr, beam_p, "Beam"))
        for constr in orientation_c:
            constraints.append(self.build_constraint(constr, orientation_p, "Crystal"))
        for constr in cell_c:
            constraints.append(self.build_constraint(constr, cell_p, "Crystal"))

        if len(constraints) == 0:
            return None

        # return constraints manager
        if options.sparse:
            return SparseConstraintManager(constraints, len(self._all_vals))
        else:
            return ConstraintManager(constraints, len(self._all_vals))



# LIBTBX_SET_DISPATCHER_NAME dials.sort_reflections


from __future__ import annotations

import dials.util
from dials.array_family import flex

help_message = """

Utility script to sort reflection tables by the values in a column.

Example::

  dials.sort_reflections key=miller_index output=sorted.refl
"""


class Sort:
    """A class for running the script."""

    def __init__(self):
        """Initialise the script."""
        from libtbx.phil import parse

        from dials.util.options import ArgumentParser

        phil_scope = parse(
            """

            key = 'miller_index'
                .type = str
                .help = "The chosen sort key. This should be a column of "
                        "the reflection table."

            reverse = False
                .type = bool
                .help = "Reverse the sort direction"

            output = sorted.refl
                .type = str
                .help = "The output reflection filename"
            """
        )

        usage = "dials.sort_reflections [options] observations.refl"

        # Initialise the base class
        self.parser = ArgumentParser(
            usage=usage, phil=phil_scope, read_reflections=True, epilog=help_message
        )

    @staticmethod
    def sort_permutation(column, reverse=False):
        indices = flex.size_t_range(len(column))
        perm = sorted(indices, key=lambda k: column[k], reverse=reverse)
        return flex.size_t(perm)

    def run(self, args=None):
        """Execute the script."""
        from dials.util.options import flatten_reflections

        # Parse the command line
        params, options = self.parser.parse_args(args, show_diff_phil=True)
        reflections = flatten_reflections(params.input.reflections)
        if not reflections:
            self.parser.print_help()
            return
        if len(reflections) != 1:
            raise dials.util.Sorry("exactly 1 reflection table must be specified")
        reflections = reflections[0]

        # Check the key is valid
        assert params.key in reflections

        # Sort the reflections
        print(f"Sorting by {params.key} with reverse={params.reverse!r}")
        perm = self.sort_permutation(reflections[params.key], params.reverse)
        reflections = reflections.select(perm)

        if options.verbose > 0:
            print("Head of sorted list " + params.key + ":")
            n = min(len(reflections), 10)
            for i in range(n):
                print(reflections[i][params.key])

        # Save sorted reflections to file
        if params.output:
            print(f"Saving reflections to {params.output}")
            reflections.as_file(params.output)


@dials.util.show_mail_handle_errors()
def run(args=None):
    script = Sort()
    script.run(args)


if __name__ == "__main__":
    run()



"""
Definition of the target function for error model minimisation.
"""


from __future__ import annotations

import numpy as np

from dxtbx import flumpy

from dials.array_family import flex


def calculate_regression_x_y(Ih_table):
    """Calculate regression data points."""
    n = Ih_table.group_multiplicities(output="per_refl")
    var_over_n_minus_1 = np.square(
        Ih_table.intensities - (Ih_table.inverse_scale_factors * Ih_table.Ih_values)
    ) / (n - 1.0)
    sigmasq_obs = Ih_table.sum_in_groups(var_over_n_minus_1, output="per_refl")
    isq = np.square(Ih_table.intensities)
    y = sigmasq_obs / isq
    x = Ih_table.variances / isq
    return x, y


class ErrorModelTarget:

    """Error model target for finding slope of norm distribution.
    (i.e. the 'a' parameter of the basic error model)"""

    _grad_names = ["d(deltahl)_dp"]
    rmsd_names = ["RMSD_deltahl"]
    rmsd_units = ["a.u"]

    def __init__(self, error_model):
        self.error_model = error_model
        # intercept + gradient
        # Quantities to cache each step
        self._rmsds = None

    def predict(self, apm):
        """Do the next step of the prediction."""
        pass

    def get_num_matches(self):
        """Get the number of reflections."""
        return self.error_model.n_refl

    def rmsds(self, apm):
        """calculate unweighted RMSDs for the matches"""
        # cache rmsd calculation for achieved test
        R = self.calculate_residuals(apm)
        n = R.size()
        R = flex.sum(R)
        R += self.compute_restraints_functional_gradients(apm)[0]
        self._rmsds = [(R / n) ** 0.5]
        return self._rmsds

    def achieved(self):
        """Method required for refinement engine."""
        return False  # implement a method here?

    # The following methods are for adaptlbfgs.
    def compute_functional_gradients(self, apm):
        """Compute the functional and gradients vector."""
        return flex.sum(self.calculate_residuals(apm)), self.calculate_gradients(apm)

    def compute_restraints_functional_gradients(self, _):
        """Compute the restraints for the functional and gradients."""
        return [0.0, 0.0]


class ErrorModelTargetRegression(ErrorModelTarget):
    def __init__(self, error_model):
        super().__init__(error_model)
        # calculate variances needed for minimisation.
        self.x, self.y = calculate_regression_x_y(self.error_model.filtered_Ih_table)
        self.n_refl = self.y.size

    def calculate_residuals(self, apm):
        """Return the residual vector"""
        params = apm.x
        if apm.active_parameters == ["a"]:
            assert len(params) == 1
            # if only a being refined, the R = y - xo*x - xo*b2
            b = self.error_model.parameters[1]
            R = (self.y - (params[0] * self.x) - params[0] * (b**2)) ** 2
        elif apm.active_parameters == ["b"]:
            assert len(params) == 1
            a = self.error_model.parameters[0]
            # R = y - a^2*x - a^2*xo
            R = (self.y - (a**2 * self.x) - (a**2 * params[0])) ** 2
        else:
            # R = y - xo*x - x1
            R = (self.y - (params[0] * self.x) - params[1]) ** 2
        return flumpy.from_numpy(R)

    def calculate_gradients(self, apm):
        "calculate the gradient vector"
        params = apm.x
        if apm.active_parameters == ["a"]:
            b = self.error_model.parameters[1]
            R = self.y - (params[0] * self.x) - params[0] * (b**2)
            gradient = flex.double([-2.0 * np.sum(R * (self.x + (b**2)))])
        elif apm.active_parameters == ["b"]:
            a = self.error_model.parameters[0]
            R = self.y - (a**2 * self.x) - (a**2 * params[0])
            gradient = flex.double([-2.0 * np.sum(R * (a**2))])
        else:
            R = self.y - (params[0] * self.x) - params[1]
            gradient = flex.double([-2.0 * np.sum(R * self.x), -2.0 * np.sum(R)])
        return gradient


class ErrorModelTargetA(ErrorModelTarget):

    """Target to minimise the 'a' component of the basic error model."""

    def calculate_residuals(self, apm):
        """Return the residual vector"""
        x = apm.x
        R = flex.pow2(
            self.error_model.sortedy - (x[1] * self.error_model.sortedx) - x[0]
        )
        return R

    def calculate_gradients(self, apm):
        "calculate the gradient vector"
        x = apm.x
        R = self.error_model.sortedy - (x[1] * self.error_model.sortedx) - x[0]
        gradient = flex.double(
            [-2.0 * flex.sum(R), -2.0 * flex.sum(R * self.error_model.sortedx)]
        )
        return gradient


class ErrorModelTargetB(ErrorModelTarget):

    """Target to minimise the 'b' component of the basic error model.

    Uses the binner to calculate residuals, gradients"""

    def predict(self, _):
        """Do the next step of the prediction."""
        self.error_model.binner.update(self.error_model.parameters)

    def calculate_residuals(self, _):
        """Return the residual vector"""
        bin_vars = self.error_model.binner.bin_variances
        R = (
            (
                np.square(np.full(bin_vars.size, 0.5) - bin_vars)
                + (1.0 / bin_vars)
                - np.full(bin_vars.size, 1.25)
            )
            * self.error_model.binner.weights
            / np.sum(self.error_model.binner.weights)
        )
        return flumpy.from_numpy(R)

    def calculate_gradients(self, apm):
        "calculate the gradient vector"
        a = self.error_model.components["a"].parameters[0]
        b = apm.x[0]
        Ih_table = self.error_model.binner.Ih_table
        I_hl = Ih_table.intensities
        g_hl = Ih_table.inverse_scale_factors
        weights = self.error_model.binner.weights
        bin_vars = self.error_model.binner.bin_variances
        sum_matrix = self.error_model.binner.summation_matrix
        bin_counts = self.error_model.binner.binning_info["refl_per_bin"]
        dsig_dc = (
            b
            * np.square(I_hl)
            * (a**2)
            / (self.error_model.binner.sigmaprime * np.square(g_hl))
        )
        ddelta_dsigma = (
            -1.0 * self.error_model.binner.delta_hl / self.error_model.binner.sigmaprime
        )
        deriv = ddelta_dsigma * dsig_dc
        dphi_by_dvar = -2.0 * (
            np.full(bin_vars.size, 0.5) - bin_vars + (1.0 / (2.0 * np.square(bin_vars)))
        )
        term1 = flumpy.to_numpy(
            flumpy.from_numpy(2.0 * self.error_model.binner.delta_hl * deriv)
            * sum_matrix
        )
        term2a = flumpy.to_numpy(
            flumpy.from_numpy(self.error_model.binner.delta_hl) * sum_matrix
        )
        term2b = flumpy.to_numpy(flumpy.from_numpy(deriv) * sum_matrix)
        grad = dphi_by_dvar * (
            (term1 / bin_counts) - (2.0 * term2a * term2b / np.square(bin_counts))
        )
        gradients = flex.double([np.sum(grad * weights) / np.sum(weights)])
        return gradients



from __future__ import annotations

import random

import pytest


def test_split_blocks_1_frame():
    from dials.algorithms.integration.integrator import JobList
    from dials.array_family import flex

    r = flex.reflection_table()
    r["value1"] = flex.double()
    r["value2"] = flex.int()
    r["value3"] = flex.double()
    r["bbox"] = flex.int6()
    r["id"] = flex.int()
    expected = []
    for i in range(100):
        x0 = random.randint(0, 100)
        x1 = x0 + random.randint(1, 10)
        y0 = random.randint(0, 100)
        y1 = y0 + random.randint(1, 10)
        z0 = random.randint(0, 100)
        z1 = z0 + random.randint(1, 10)
        v1 = random.uniform(0, 100)
        v2 = random.randint(0, 100)
        v3 = random.uniform(0, 100)
        r.append(
            {
                "id": 0,
                "value1": v1,
                "value2": v2,
                "value3": v3,
                "bbox": (x0, x1, y0, y1, z0, z1),
            }
        )
        for z in range(z0, z1):
            expected.append(
                {
                    "id": 0,
                    "value1": v1,
                    "value2": v2,
                    "value3": v3,
                    "bbox": (x0, x1, y0, y1, z, z + 1),
                    "partial_id": i,
                }
            )

    jobs = JobList()
    jobs.add((0, 1), (0, 111), 1, 0)

    jobs.split(r)
    assert len(r) == len(expected)
    EPS = 1e-7
    for r1, r2 in zip(r.rows(), expected):
        assert r1["bbox"] == r2["bbox"]
        assert r1["partial_id"] == r2["partial_id"]
        assert abs(r1["value1"] - r2["value1"]) < EPS
        assert r1["value2"] == r2["value2"]
        assert abs(r1["value3"] - r2["value3"]) < EPS


def test_split_blocks_non_overlapping():
    from scitbx.array_family import shared

    from dials.algorithms.integration.integrator import JobList
    from dials.array_family import flex

    blocks = shared.tiny_int_2(
        [
            (0, 10),
            (10, 20),
            (20, 30),
            (30, 35),
            (35, 40),
            (40, 50),
            (50, 60),
            (60, 70),
            (70, 80),
            (80, 90),
            (90, 100),
            (100, 110),
        ]
    )

    jobs = JobList((0, 1), blocks)

    r = flex.reflection_table()
    r["value1"] = flex.double()
    r["value2"] = flex.int()
    r["value3"] = flex.double()
    r["bbox"] = flex.int6()
    r["id"] = flex.int()
    expected = []
    for i in range(100):
        x0 = random.randint(0, 100)
        x1 = x0 + random.randint(1, 10)
        y0 = random.randint(0, 100)
        y1 = y0 + random.randint(1, 10)
        z0 = random.randint(0, 100)
        z1 = z0 + random.randint(1, 10)
        v1 = random.uniform(0, 100)
        v2 = random.randint(0, 100)
        v3 = random.uniform(0, 100)
        r.append(
            {
                "id": 0,
                "value1": v1,
                "value2": v2,
                "value3": v3,
                "bbox": (x0, x1, y0, y1, z0, z1),
            }
        )

        for j in range(len(blocks)):
            b0 = blocks[j][0]
            b1 = blocks[j][1]
            if (
                (z0 >= b0 and z1 <= b1)
                or (z0 < b1 and z1 >= b1)
                or (z0 < b0 and z1 > b0)
            ):
                z00 = max(b0, z0)
                z11 = min(b1, z1)
                expected.append(
                    {
                        "id": 0,
                        "value1": v1,
                        "value2": v2,
                        "value3": v3,
                        "bbox": (x0, x1, y0, y1, z00, z11),
                        "partial_id": i,
                    }
                )

    jobs.split(r)
    assert len(r) == len(expected)
    EPS = 1e-7
    for r1, r2 in zip(r.rows(), expected):
        assert r1["bbox"] == r2["bbox"]
        assert r1["partial_id"] == r2["partial_id"]
        assert abs(r1["value1"] - r2["value1"]) < EPS
        assert r1["value2"] == r2["value2"]
        assert abs(r1["value3"] - r2["value3"]) < EPS


def test_split_blocks_overlapping():
    from scitbx.array_family import shared

    from dials.algorithms.integration.integrator import JobList
    from dials.array_family import flex

    blocks = shared.tiny_int_2(
        [
            (0, 10),
            (5, 15),
            (10, 20),
            (15, 25),
            (20, 30),
            (25, 35),
            (30, 40),
            (35, 45),
            (40, 50),
            (45, 55),
            (50, 60),
            (55, 65),
            (60, 70),
            (65, 75),
            (70, 80),
            (75, 85),
            (80, 90),
            (85, 95),
            (90, 100),
            (95, 105),
            (100, 110),
        ]
    )

    jobs = JobList((0, 1), blocks)

    r = flex.reflection_table()
    r["value1"] = flex.double()
    r["value2"] = flex.int()
    r["value3"] = flex.double()
    r["bbox"] = flex.int6()
    r["id"] = flex.int()
    expected = []
    for i in range(100):
        x0 = random.randint(0, 100)
        x1 = x0 + random.randint(1, 10)
        y0 = random.randint(0, 100)
        y1 = y0 + random.randint(1, 10)
        z0 = random.randint(0, 90)
        z1 = z0 + random.randint(1, 20)
        v1 = random.uniform(0, 100)
        v2 = random.randint(0, 100)
        v3 = random.uniform(0, 100)
        r.append(
            {
                "id": 0,
                "value1": v1,
                "value2": v2,
                "value3": v3,
                "bbox": (x0, x1, y0, y1, z0, z1),
            }
        )
        expected.append(
            {
                "id": 0,
                "value1": v1,
                "value2": v2,
                "value3": v3,
                "bbox": (x0, x1, y0, y1, z0, z1),
            }
        )

    jobs.split(r)
    assert len(r) > 100
    for r1 in r.rows():
        v1 = r1["value1"]
        v2 = r1["value2"]
        v3 = r1["value3"]
        bbox = r1["bbox"]
        pid = r1["partial_id"]

        z0 = bbox[4]
        z1 = bbox[5]
        success = False
        for i in range(len(blocks)):
            b0 = blocks[i][0]
            b1 = blocks[i][1]
            if z0 >= b0 and z1 <= b1:
                success = True
                break
        assert success

        v11 = expected[pid]["value1"]
        v22 = expected[pid]["value2"]
        v33 = expected[pid]["value3"]
        bb = expected[pid]["bbox"]
        assert v11 == v1
        assert v22 == v2
        assert v33 == v3
        assert bb[0] == bbox[0]
        assert bb[1] == bbox[1]
        assert bb[2] == bbox[2]
        assert bb[3] == bbox[3]


def test_reflection_manager():
    from dials.array_family import flex

    reflections = flex.reflection_table()
    reflections["panel"] = flex.size_t()
    reflections["bbox"] = flex.int6()
    reflections["miller_index"] = flex.miller_index()
    reflections["s1"] = flex.vec3_double()
    reflections["xyzcal.px"] = flex.vec3_double()
    reflections["xyzcal.mm"] = flex.vec3_double()
    reflections["entering"] = flex.bool()
    reflections["id"] = flex.int()
    reflections["flags"] = flex.size_t()

    width = 1000
    height = 1000
    nrefl = 10000
    array_range = (0, 130)
    block_size = 20
    block_overlap = 10

    random.seed(0)
    processed = [[] for i in range(12)]
    for i in range(nrefl):
        x0 = random.randint(0, width - 10)
        y0 = random.randint(0, height - 10)
        zs = random.randint(2, 9)
        x1 = x0 + random.randint(2, 10)
        y1 = y0 + random.randint(2, 10)
        for k, j in enumerate([10, 20, 30, 40, 50, 60, 70, 80, 90, 100, 110, 120]):
            m = k + i * 12
            pos = random.choice(["left", "right", "centre"])
            if pos == "left":
                z0 = j - zs
                z1 = j
            elif pos == "right":
                z0 = j
                z1 = j + zs
            else:
                z0 = j - zs // 2
                z1 = j + zs // 2
            bbox = (x0, x1, y0, y1, z0, z1)
            reflections.append(
                {
                    "panel": random.randint(0, 1),
                    "bbox": bbox,
                    "flags": flex.reflection_table.flags.reference_spot,
                }
            )
            processed[k].append(m)

        # Add reflection to ignore
        # zc = random.choice([10, 20, 30, 40, 50, 60, 70, 80, 90, 100, 110, 120])
        # z0 = zc - 11
        # z1 = zc + 11
        # bbox = (x0, x1, y0, y1, z0, z1)
        # reflections.append({
        #   "panel" : randint(0,1),
        #   "bbox" : bbox,
        #   "flags" : flex.reflection_table.flags.reference_spot
        # })

    from dials.algorithms.integration.integrator import JobList, ReflectionManager

    jobs = JobList()
    jobs.add((0, 1), array_range, block_size, block_overlap)

    # Create the executor
    executor = ReflectionManager(jobs, reflections)

    # Ensure the tasks make sense
    jobs = [executor.job(i) for i in range(len(executor))]
    assert len(executor) == 12
    assert not executor.finished()
    assert len(jobs) == 12
    assert jobs[0].frames() == (0, 20)
    assert jobs[1].frames() == (10, 30)
    assert jobs[2].frames() == (20, 40)
    assert jobs[3].frames() == (30, 50)
    assert jobs[4].frames() == (40, 60)
    assert jobs[5].frames() == (50, 70)
    assert jobs[6].frames() == (60, 80)
    assert jobs[7].frames() == (70, 90)
    assert jobs[8].frames() == (80, 100)
    assert jobs[9].frames() == (90, 110)
    assert jobs[10].frames() == (100, 120)
    assert jobs[11].frames() == (110, 130)

    # Get the task specs
    data0 = executor.split(0)
    data1 = executor.split(1)
    data2 = executor.split(2)
    data3 = executor.split(3)
    data4 = executor.split(4)
    data5 = executor.split(5)
    data6 = executor.split(6)
    data7 = executor.split(7)
    data8 = executor.split(8)
    data9 = executor.split(9)
    data10 = executor.split(10)
    data11 = executor.split(11)
    assert len(data0) == len(processed[0])
    assert len(data1) == len(processed[1])
    assert len(data2) == len(processed[2])
    assert len(data3) == len(processed[3])
    assert len(data4) == len(processed[4])
    assert len(data5) == len(processed[5])
    assert len(data6) == len(processed[6])
    assert len(data7) == len(processed[7])
    assert len(data8) == len(processed[8])
    assert len(data9) == len(processed[9])
    assert len(data10) == len(processed[10])
    assert len(data11) == len(processed[11])

    # Add some results
    data0["data"] = flex.double(len(data0), 1)
    data1["data"] = flex.double(len(data1), 2)
    data2["data"] = flex.double(len(data2), 3)
    data3["data"] = flex.double(len(data3), 4)
    data4["data"] = flex.double(len(data4), 5)
    data5["data"] = flex.double(len(data5), 6)
    data6["data"] = flex.double(len(data6), 7)
    data7["data"] = flex.double(len(data7), 8)
    data8["data"] = flex.double(len(data8), 9)
    data9["data"] = flex.double(len(data9), 10)
    data10["data"] = flex.double(len(data10), 11)
    data11["data"] = flex.double(len(data11), 12)

    # Accumulate the data again
    assert not executor.finished()
    executor.accumulate(0, data0)
    executor.accumulate(1, data1)
    executor.accumulate(2, data2)
    executor.accumulate(3, data3)
    executor.accumulate(4, data4)
    executor.accumulate(5, data5)
    executor.accumulate(6, data6)
    executor.accumulate(7, data7)
    executor.accumulate(8, data8)
    executor.accumulate(9, data9)
    executor.accumulate(10, data10)
    executor.accumulate(11, data11)
    assert executor.finished()

    # Get results and check they're as expected
    data = executor.data()
    result = data["data"]
    for i in range(len(processed)):
        for j in range(len(processed[i])):
            assert result[processed[i][j]] == i + 1

    # Test passed


@pytest.mark.parametrize("nproc", [1, 2])
def test_integrator_3d(dials_data, nproc):
    from math import pi

    from dxtbx.model.experiment_list import ExperimentListFactory

    from dials.algorithms.profile_model.gaussian_rs import Model
    from dials.array_family import flex

    path = dials_data("centroid_test_data", pathlib=True) / "experiments.json"

    exlist = ExperimentListFactory.from_json_file(path)
    exlist[0].profile = Model(
        None, n_sigma=3, sigma_b=0.024 * pi / 180.0, sigma_m=0.044 * pi / 180.0
    )

    rlist = flex.reflection_table.from_predictions(exlist[0])
    rlist["id"] = flex.int(len(rlist), 0)
    rlist.compute_bbox(exlist)
    rlist.compute_zeta_multi(exlist)
    rlist.compute_d(exlist)

    from libtbx.phil import parse

    from dials.algorithms.integration.integrator import Integrator3D, phil_scope

    params = phil_scope.fetch(
        parse(
            """
    integration.block.size=%d
    integration.mp.nproc=%d
    integration.profile_fitting=False
  """
            % (5, nproc)
        )
    ).extract()

    integrator = Integrator3D(exlist, rlist, params)
    integrator.integrate()


def test_summation(dials_data):
    from math import pi

    from dxtbx.model.experiment_list import ExperimentListFactory

    from dials.algorithms.profile_model.gaussian_rs import Model
    from dials.array_family import flex

    path = dials_data("centroid_test_data", pathlib=True) / "experiments.json"

    exlist = ExperimentListFactory.from_json_file(path)
    exlist[0].profile = Model(
        None, n_sigma=3, sigma_b=0.024 * pi / 180.0, sigma_m=0.044 * pi / 180.0
    )

    rlist = flex.reflection_table.from_predictions(exlist[0])
    rlist["id"] = flex.int(len(rlist), 0)

    def integrate(integrator_type, rlist):
        from libtbx.phil import parse

        from dials.algorithms.integration.integrator import create_integrator
        from dials.algorithms.integration.integrator import (
            phil_scope as master_phil_scope,
        )

        rlist = rlist.copy()

        phil_scope = parse(
            f"""
      integration.background.algorithm=null
      integration.intensity.algorithm=sum
      integration.intensity.sum.integrator={integrator_type}
      integration.block.size=0.5
      integration.profile_fitting=False
    """
        )

        params = master_phil_scope.fetch(source=phil_scope).extract()

        integrator = create_integrator(params, exlist, rlist)

        result = integrator.integrate()
        return result

    from libtbx.test_utils import approx_equal

    def approx_equal_dict(a, b, k):
        return approx_equal(a[k], b[k])

    # Do summation by all different methods
    result1 = integrate("3d", rlist)
    result2 = integrate("flat3d", rlist)
    result3 = integrate("2d", rlist)
    result4 = integrate("single2d", rlist)
    assert len(result1) >= len(rlist)
    assert len(result2) >= len(rlist)
    assert len(result3) >= len(rlist)
    assert len(result4) >= len(rlist)

    # result1 and result2 should be the same
    assert len(result1) == len(result2)
    for r1, r2 in zip(result1.rows(), result2.rows()):
        assert r1["partial_id"] == r2["partial_id"]
        assert r1["bbox"] == r2["bbox"]
        assert r1["entering"] == r2["entering"]
        assert r1["flags"] == r2["flags"]
        assert r1["id"] == r2["id"]
        assert r1["miller_index"] == r2["miller_index"]
        assert r1["panel"] == r2["panel"]
        assert approx_equal_dict(r1, r2, "d")
        assert approx_equal_dict(r1, r2, "intensity.sum.value")
        assert approx_equal_dict(r1, r2, "intensity.sum.variance")
        assert approx_equal_dict(r1, r2, "lp")
        assert approx_equal_dict(r1, r2, "partiality")
        assert approx_equal_dict(r1, r2, "s1")
        assert approx_equal_dict(r1, r2, "xyzcal.mm")
        assert approx_equal_dict(r1, r2, "xyzcal.px")
        assert approx_equal_dict(r1, r2, "zeta")

    # result3 and result4 should be the same
    assert len(result3) == len(result4)
    for r3, r4 in zip(result3.rows(), result4.rows()):
        assert r3["partial_id"] == r4["partial_id"]
        assert r3["bbox"] == r4["bbox"]
        assert r3["entering"] == r4["entering"]
        assert r3["flags"] == r4["flags"]
        assert r3["id"] == r4["id"]
        assert r3["miller_index"] == r4["miller_index"]
        assert r3["panel"] == r4["panel"]
        assert approx_equal_dict(r3, r4, "d")
        assert approx_equal_dict(r3, r4, "intensity.sum.value")
        assert approx_equal_dict(r3, r4, "intensity.sum.variance")
        assert approx_equal_dict(r3, r4, "lp")
        assert approx_equal_dict(r3, r4, "partiality")
        assert approx_equal_dict(r3, r4, "s1")
        assert approx_equal_dict(r3, r4, "xyzcal.mm")
        assert approx_equal_dict(r3, r4, "xyzcal.px")
        assert approx_equal_dict(r3, r4, "xyzobs.px.value")
        assert approx_equal_dict(r3, r4, "xyzobs.px.variance")
        assert approx_equal_dict(r3, r4, "zeta")

    # result3 should add up to result1
    assert len(result3) >= len(result1)
    expected1 = rlist.copy()
    expected1["intensity.sum.value"] = flex.double(len(rlist), 0)
    expected1["intensity.sum.variance"] = flex.double(len(rlist), 0)
    for r1 in result1.rows():
        pid = r1["partial_id"]
        r2 = expected1[pid]
        assert r1["entering"] == r2["entering"]
        assert r1["id"] == r2["id"]
        assert r1["miller_index"] == r2["miller_index"]
        assert r1["panel"] == r2["panel"]
        assert approx_equal_dict(r1, r2, "s1")
        assert approx_equal_dict(r1, r2, "xyzcal.mm")
        assert approx_equal_dict(r1, r2, "xyzcal.px")
        expected1["intensity.sum.value"][pid] += r1["intensity.sum.value"]
        expected1["intensity.sum.variance"][pid] += r1["intensity.sum.variance"]
    expected3 = rlist.copy()
    expected3["intensity.sum.value"] = flex.double(len(rlist), 0)
    expected3["intensity.sum.variance"] = flex.double(len(rlist), 0)
    for r1 in result3.rows():
        pid = r1["partial_id"]
        r2 = expected3[pid]
        assert r1["entering"] == r2["entering"]
        assert r1["id"] == r2["id"]
        assert r1["miller_index"] == r2["miller_index"]
        assert r1["panel"] == r2["panel"]
        assert approx_equal_dict(r1, r2, "s1")
        assert approx_equal_dict(r1, r2, "xyzcal.mm")
        assert approx_equal_dict(r1, r2, "xyzcal.px")
        expected3["intensity.sum.value"][pid] += r1["intensity.sum.value"]
        expected3["intensity.sum.variance"][pid] += r1["intensity.sum.variance"]
    for r1, r3 in zip(expected1.rows(), expected3.rows()):
        assert approx_equal_dict(r1, r3, "intensity.sum.value")
        assert approx_equal_dict(r1, r3, "intensity.sum.variance")



from __future__ import annotations

import copy
import logging
import math
import time
from collections import namedtuple
from typing import Tuple

import libtbx.phil
from cctbx import crystal
from dxtbx.masking import (
    mask_untrusted_circle,
    mask_untrusted_polygon,
    mask_untrusted_rectangle,
)
from dxtbx.model import ImageSet, SimplePxMmStrategy
from iotbx.phil import parse

from dials.array_family import flex
from dials.util.ext import ResolutionMaskGenerator

logger = logging.getLogger(__name__)

phil_scope = parse(
    """
  border = 0
    .type = int
    .help = "The border around the edge of the image."

  d_min = None
    .help = "The high resolution limit in Angstrom for a pixel to be"
            "accepted by the filtering algorithm."
    .type = float(value_min=0)

  d_max = None
    .help = "The low resolution limit in Angstrom for a pixel to be"
            "accepted by the filtering algorithm."
    .type = float(value_min=0)

  disable_parallax_correction = False
    .help = "Set to ``True`` to use a faster, but less accurate, simple px-to-mm "
            "mapping by disabling accounting for parallax correction when generating "
            "resolution masks."
    .type = bool

  resolution_range = None
    .multiple = true
    .type = floats(2)
    .help = "an untrusted resolution range"

  untrusted
    .multiple = True
  {

    panel = None
      .type = int
      .help = "The panel number"
      .help = "If no geometric attributes are given (circle, rectangle, etc)"
      .help = "then the whole panel is masked out"

    circle = None
      .type = ints(3)
      .help = "An untrusted circle (xc, yc, r)"

    rectangle = None
      .type = ints(4)
      .help = "An untrusted rectangle (x0, x1, y0, y1)"

    polygon = None
      .type = ints(value_min=0)
      .help = "The pixel coordinates (fast, slow) that define the corners "
              "of the untrusted polygon. Spots whose centroids fall within "
              "the bounds of the untrusted polygon will be rejected."

    pixel = None
      .type = ints(2, value_min=0)
      .help = "An untrusted pixel (y, x)"

  }

  ice_rings {
    filter = False
      .type = bool
    unit_cell = 4.498,4.498,7.338,90,90,120
      .type = unit_cell
      .help = "The unit cell to generate d_spacings for powder rings."
      .expert_level = 1
    space_group = 194
      .type = space_group
      .help = "The space group used to generate d_spacings for powder rings."
      .expert_level = 1
    width = 0.002
      .type = float(value_min=0.0)
      .help = "The width of an ice ring (in 1/d^2)."
      .expert_level = 1
    d_min = None
      .type = float(value_min=0.0)
      .help = "The high resolution limit (otherwise use detector d_min)"
      .expert_level = 1
  }
""",
    process_includes=True,
)

CacheInfo = namedtuple("CacheInfo", ["hits", "misses", "maxsize", "currsize"])


def lru_equality_cache(maxsize=10):
    """LRU cache that compares keys based on equality... inefficiently.

    Used for dxtbx models that don't have unique id values so can't be
    cached with a normal lru_cache.

    Args:
        maxsize (int): The maximum number of old results to remember
    """

    def _decorator(f):
        class Scope:
            pass

        cache_data = Scope()
        cache_data.cache = []
        cache_data.hits = 0
        cache_data.misses = 0

        def _wrapper_function(*args, **kwargs):
            for i, (key_args, key_kwargs, key_result) in enumerate(cache_data.cache):
                if key_args == args and key_kwargs == kwargs:
                    cache_data.cache.append(cache_data.cache.pop(i))
                    cache_data.hits += 1
                    return key_result
            result = f(*args, **kwargs)
            cache_data.misses += 1
            cache_data.cache.append((args, kwargs, result))
            if len(cache_data.cache) > maxsize:
                cache_data.cache = cache_data.cache[1:]
            return result

        def _generate_cache_info():
            return CacheInfo(
                hits=cache_data.hits,
                misses=cache_data.misses,
                maxsize=maxsize,
                currsize=len(cache_data.cache),
            )

        _wrapper_function.__wrapped__ = f
        _wrapper_function.cache_info = _generate_cache_info

        return _wrapper_function

    return _decorator


def generate_ice_ring_resolution_ranges(beam, panel, params):
    """
    Generate a set of resolution ranges from the ice ring parameters
    """
    if params.filter is True:

        # Get the crystal symmetry
        crystal_symmetry = crystal.symmetry(
            unit_cell=params.unit_cell, space_group=params.space_group.group()
        )

        # Get the half width
        half_width = params.width * 0.5

        # Set the high resolution
        if params.d_min is None:
            d_min = panel.get_max_resolution_at_corners(beam.get_s0())
        else:
            d_min = params.d_min

        # Build the miller set
        ms = crystal_symmetry.build_miller_set(anomalous_flag=False, d_min=d_min)
        ms = ms.sort(by_value="resolution")

        # Yield all the d ranges
        for j, d in enumerate(ms.d_spacings().data()):
            d_sq_inv = 1.0 / (d**2)
            d_sq_inv_min = d_sq_inv - half_width
            d_sq_inv_max = d_sq_inv + half_width
            d_max = math.sqrt(1.0 / d_sq_inv_min)
            d_min = math.sqrt(1.0 / d_sq_inv_max)
            yield (d_min, d_max)


@lru_equality_cache(maxsize=3)
def _get_resolution_masker(beam, panel):
    t0 = time.perf_counter()
    masker = ResolutionMaskGenerator(beam, panel)
    t1 = time.perf_counter()
    logger.debug(f"ResolutionMaskGenerator calculation took {t1 - t0:.4f} seconds")
    return masker


def _apply_resolution_mask(mask, beam, panel, *args):
    _get_resolution_masker(beam, panel).apply(mask, *args)


def generate_mask(
    imageset: ImageSet, params: libtbx.phil.scope_extract
) -> Tuple[flex.bool]:
    """Generate a mask based on the input parameters.

    Args:
      imageset (ImageSet): The imageset for which to generate a mask
      params (libtbx.phil.scope_extract): The phil parameters for mask generation. This
        should be an extract of `dials.util.masking.phil_scope`
    """
    # Get the detector and beam
    detector = imageset.get_detector()
    beam = imageset.get_beam()

    # Create the mask for each panel
    masks = []
    for index, panel in enumerate(detector):

        mask = flex.bool(flex.grid(reversed(panel.get_image_size())), True)

        # Add a border around the image
        if params.border > 0:
            logger.debug(f"Generating border mask:\n border = {params.border}")
            border = params.border
            height, width = mask.all()
            borderx = flex.bool(flex.grid(border, width), False)
            bordery = flex.bool(flex.grid(height, border), False)
            mask[0:border, :] = borderx
            mask[-border:, :] = borderx
            mask[:, 0:border] = bordery
            mask[:, -border:] = bordery

        # Apply the untrusted regions
        for region in params.untrusted:
            if region.panel is None:
                region.panel = 0
            if region.panel == index:
                if not any(
                    [region.circle, region.rectangle, region.polygon, region.pixel]
                ):
                    mask[:, :] = flex.bool(flex.grid(mask.focus()), False)
                    continue

                if region.circle is not None:
                    xc, yc, radius = region.circle
                    logger.debug(
                        "Generating circle mask:\n"
                        + f" panel = {region.panel}\n"
                        + f" xc = {xc}\n"
                        + f" yc = {yc}\n"
                        + f" radius = {radius}"
                    )
                    mask_untrusted_circle(mask, xc, yc, radius)
                if region.rectangle is not None:
                    x0, x1, y0, y1 = region.rectangle
                    logger.debug(
                        "Generating rectangle mask:\n"
                        + f" panel = {region.panel}\n"
                        + f" x0 = {x0}\n"
                        + f" y0 = {y0}\n"
                        + f" x1 = {x1}\n"
                        + f" y1 = {y1}"
                    )
                    mask_untrusted_rectangle(mask, x0, x1, y0, y1)
                if region.polygon is not None:
                    assert (
                        len(region.polygon) % 2 == 0
                    ), "Polygon must contain 2D coords"
                    vertices = []
                    for i in range(int(len(region.polygon) / 2)):
                        x = region.polygon[2 * i]
                        y = region.polygon[2 * i + 1]
                        vertices.append((x, y))
                    polygon = flex.vec2_double(vertices)
                    logger.debug(
                        f"Generating polygon mask:\n panel = {region.panel}\n"
                        + "\n".join(f" coord = {vertex}" for vertex in vertices)
                    )
                    mask_untrusted_polygon(mask, polygon)
                if region.pixel is not None:
                    mask[region.pixel] = False

        # PxMmStrategy to use for generating resolution masks
        if params.disable_parallax_correction:
            panel = copy.deepcopy(panel)
            panel.set_px_mm_strategy(SimplePxMmStrategy())

        # Generate high and low resolution masks
        if params.d_min is not None:
            logger.debug(f"Generating high resolution mask:\n d_min = {params.d_min}")
            _apply_resolution_mask(mask, beam, panel, 0, params.d_min)
        if params.d_max is not None:
            logger.debug(f"Generating low resolution mask:\n d_max = {params.d_max}")
            d_max = params.d_max
            d_inf = max(d_max + 1, 1e9)
            _apply_resolution_mask(mask, beam, panel, d_max, d_inf)

        try:
            # Mask out the resolution range
            for drange in params.resolution_range:
                d_min = min(drange)
                d_max = max(drange)
                assert d_min < d_max, "d_min must be < d_max"
                logger.debug(
                    "Generating resolution range mask:\n"
                    + f" d_min = {d_min}\n"
                    + f" d_max = {d_max}"
                )
                _apply_resolution_mask(mask, beam, panel, d_min, d_max)
        except TypeError:
            # Catch the default value None of params.resolution_range
            if any(params.resolution_range):
                raise

        # Mask out the resolution ranges for the ice rings
        for drange in generate_ice_ring_resolution_ranges(
            beam, panel, params.ice_rings
        ):
            d_min = min(drange)
            d_max = max(drange)
            assert d_min < d_max, "d_min must be < d_max"
            logger.debug(
                "Generating ice ring mask:\n"
                + f" d_min = {d_min:.4f}\n"
                + f" d_max = {d_max:.4f}"
            )
            _apply_resolution_mask(mask, beam, panel, d_min, d_max)

        # Add to the list
        masks.append(mask)

    # Return the mask
    return tuple(masks)



"""
This file defines targets for scaling.

These are initialised with a scaler and an active parameter manager,
and have implementations of residual/gradient calculations for
scaling.
"""

from __future__ import annotations

import numpy as np

from dxtbx import flumpy

from dials.algorithms.scaling.scaling_restraints import ScalingRestraintsCalculator
from dials.array_family import flex
from dials_scaling_ext import calc_dIh_by_dpi, calc_jacobian, row_multiply


class ScalingTarget:
    """
    A class to be used by a Scaling Refinery to calculate gradients,
    residuals etc required by the Refinery for minimisation.
    '"""

    _grad_names = ["dI_dp"]
    rmsd_names = ["RMSD_I"]
    rmsd_units = ["a.u"]

    def __init__(self):
        self.rmsd_names = ["RMSD_I"]
        self.rmsd_units = ["a.u"]
        # Quantities to cache each step
        self._rmsds = None
        self.param_restraints = True  # If one tests for restraints and None is
        # returned, then this is set to False and restraints calculations are not
        # attempted for the remainder of the minimisation with this target function.

    def rmsds(self, Ih_table, apm):
        """Calculate RMSDs for the matches. Also calculate R-factors."""
        R = 0
        n = 0
        for block in Ih_table.blocked_data_list:
            R += np.sum(np.square(self.calculate_residuals(block)) * block.weights)
            n += block.size
        if self.param_restraints:
            restraints = ScalingRestraintsCalculator.calculate_restraints(apm)
            if restraints:
                R += np.sum(restraints[0])
            else:
                self.param_restraints = False
        self._rmsds = [(R / n) ** 0.5]
        return self._rmsds

    @staticmethod
    def achieved():
        """Method required by refinement engine."""
        return False  # implement a method here?

    @staticmethod
    def calculate_residuals(Ih_table):
        """Return the residual vector."""
        R = Ih_table.intensities - (Ih_table.inverse_scale_factors * Ih_table.Ih_values)
        return R

    @staticmethod
    def calculate_gradients(Ih_table):
        """Return a gradient vector on length len(self.apm.x)."""
        gsq = np.square(Ih_table.inverse_scale_factors) * Ih_table.weights
        sumgsq = flumpy.from_numpy(Ih_table.sum_in_groups(gsq))
        prefactor = (
            -2.0
            * Ih_table.weights
            * (
                Ih_table.intensities
                - (Ih_table.Ih_values * Ih_table.inverse_scale_factors)
            )
        )
        dIh = flumpy.from_numpy(
            (
                Ih_table.intensities
                - (Ih_table.Ih_values * 2.0 * Ih_table.inverse_scale_factors)
            )
            * Ih_table.weights
        )
        dIh_by_dpi = calc_dIh_by_dpi(
            dIh, sumgsq, Ih_table.h_index_matrix, Ih_table.derivatives.transpose()
        )
        term_1 = (
            flumpy.from_numpy(prefactor * Ih_table.Ih_values) * Ih_table.derivatives
        )
        term_2 = (
            flumpy.from_numpy(
                Ih_table.sum_in_groups(prefactor * Ih_table.inverse_scale_factors)
            )
            * dIh_by_dpi
        )
        gradient = term_1 + term_2
        return gradient

    @staticmethod
    def calculate_jacobian(Ih_table):
        """Calculate the jacobian matrix, size Ih_table.size by len(self.apm.x)."""
        gsq = np.square(Ih_table.inverse_scale_factors) * Ih_table.weights
        sumgsq = Ih_table.sum_in_groups(gsq)
        dIh = (
            Ih_table.intensities
            - (Ih_table.Ih_values * 2.0 * Ih_table.inverse_scale_factors)
        ) * Ih_table.weights
        jacobian = calc_jacobian(
            Ih_table.derivatives.transpose(),
            Ih_table.h_index_matrix,
            flumpy.from_numpy(Ih_table.Ih_values),
            flumpy.from_numpy(Ih_table.inverse_scale_factors),
            flumpy.from_numpy(dIh),
            flumpy.from_numpy(sumgsq),
        )
        return jacobian

    # The following methods are for adaptlbfgs.
    @classmethod
    def compute_functional_gradients(cls, Ih_table):
        """Return the functional and gradients."""
        resids = cls.calculate_residuals(Ih_table)
        gradients = cls.calculate_gradients(Ih_table)
        weights = Ih_table.weights
        functional = np.sum(np.square(resids) * weights)
        del Ih_table.derivatives
        return functional, gradients

    def compute_restraints_functional_gradients(self, apm):
        """Return the restrains for functional and gradients."""
        restraints = None
        if self.param_restraints:
            restr = ScalingRestraintsCalculator.calculate_restraints(apm)
            if restr:
                resid_restr = flex.sum(restr[0])  # add to total functional here
                grad_restr = restr[1]
                restraints = [resid_restr, grad_restr]
            else:
                self.param_restraints = False
        return restraints  # list of restraints to add to resid, grads and curvs

    # The following methods are for adaptlstbx (GN/ LM algorithms)
    @classmethod
    def compute_residuals(cls, Ih_table):
        """Return the residuals array and weights."""
        residuals = cls.calculate_residuals(Ih_table)
        weights = Ih_table.weights
        return flumpy.from_numpy(residuals), flumpy.from_numpy(weights)

    @classmethod
    def compute_residuals_and_gradients(cls, Ih_table):
        """Return the residuals array, jacobian matrix and weights."""
        residuals = cls.calculate_residuals(Ih_table)
        jacobian = cls.calculate_jacobian(Ih_table)
        weights = Ih_table.weights
        Ih_table.derivatives = None
        return flumpy.from_numpy(residuals), jacobian, flumpy.from_numpy(weights)

    def compute_restraints_residuals_and_gradients(self, apm):
        """Return the restraints for the residuals and jacobian."""
        if self.param_restraints:
            restr = ScalingRestraintsCalculator.calculate_jacobian_restraints(apm)
            if not restr:
                self.param_restraints = False
            return restr
        return None


class ScalingTargetFixedIH(ScalingTarget):
    """An implementation of scaling target for when the scaling is to be
    done against a fixed reference Ih set (i.e scaler is a TargetScaler)
    """

    @staticmethod
    def calculate_gradients(Ih_table):
        rhl = Ih_table.intensities - (
            Ih_table.Ih_values * Ih_table.inverse_scale_factors
        )
        G = (
            flumpy.from_numpy(-2.0 * rhl * Ih_table.weights * Ih_table.Ih_values)
            * Ih_table.derivatives
        )
        return G

    @staticmethod
    def calculate_jacobian(Ih_table):
        """Calculate the jacobian matrix, size Ih_table.size by len(self.apm.x)."""
        jacobian = row_multiply(
            Ih_table.derivatives, flumpy.from_numpy(-1.0 * Ih_table.Ih_values)
        )
        return jacobian



"""Test that compute_delta_cchalf returns required values"""
from __future__ import annotations

import json
from unittest import mock

from dxtbx.model import Crystal, Experiment, Scan
from dxtbx.model.experiment_list import ExperimentList
from libtbx import phil

from dials.algorithms.scaling.model.model import KBScalingModel
from dials.algorithms.scaling.scale_and_filter import AnalysisResults, log_cycle_results
from dials.algorithms.statistics.cc_half_algorithm import CCHalfFromDials, DeltaCCHalf
from dials.array_family import flex
from dials.util.options import ArgumentParser


def generate_test_reflections(n=2):
    """Generate data for testing."""
    reflections = flex.reflection_table()
    for id_ in range(0, n):
        r = flex.reflection_table()
        r["id"] = flex.int(10, id_)
        r["xyzobs.px.value"] = flex.vec3_double([(0, 0, i + 0.5) for i in range(0, 10)])
        r.experiment_identifiers()[id_] = str(id_)
        r.set_flags(flex.bool(10, True), r.flags.integrated)
        r.set_flags(flex.bool(10, True), r.flags.scaled)
        reflections.extend(r)
    return reflections


def generated_params():
    """Generate a param phil scope."""
    phil_scope = phil.parse(
        """
      include scope dials.algorithms.scaling.scaling_options.phil_scope
      include scope dials.algorithms.scaling.model.model.model_phil_scope
      include scope dials.algorithms.scaling.scaling_refiner.scaling_refinery_phil_scope
  """,
        process_includes=True,
    )
    parser = ArgumentParser(phil=phil_scope, check_format=False)
    parameters, _ = parser.parse_args(args=[], quick_parse=True, show_diff_phil=False)
    parameters.model = "KB"
    return parameters


def get_scaling_model():
    """Make a KB Scaling model instance"""
    return KBScalingModel.from_data(generated_params(), [], [])


def generate_test_experiments(n=2):
    """Make a test experiment list"""
    experiments = ExperimentList()
    exp_dict = {
        "__id__": "crystal",
        "real_space_a": [1.0, 0.0, 0.0],
        "real_space_b": [0.0, 1.0, 0.0],
        "real_space_c": [0.0, 0.0, 2.0],
        "space_group_hall_symbol": " C 2y",
    }
    for i in range(n):
        experiments.append(
            Experiment(
                crystal=Crystal.from_dict(exp_dict),
                scan=Scan(image_range=[1, 10], oscillation=[0.0, 1.0]),
                scaling_model=get_scaling_model(),
                identifier=str(i),
            )
        )
    return experiments


def test_scale_and_filter_results_logging():
    """Test ScaleAndFilter.log_cycle_results method."""
    results = AnalysisResults()
    scaling_script = mock.Mock()
    scaling_script.merging_statistics_result = "stats_results"
    scaling_script.scaled_miller_array.size.return_value = 1000

    filter_script = mock.Mock()
    filter_script.results_summary = {
        "dataset_removal": {
            "mode": "image_group",
            "image_ranges_removed": [[(6, 10), 0]],
            "experiments_fully_removed": [],
            "experiment_ids_fully_removed": [],
            "n_reflections_removed": 50,
        },
        "mean_cc_half": 80.0,
        "per_dataset_delta_cc_half_values": {
            "delta_cc_half_values": [-0.1, 0.1, -0.2, 0.2]
        },
    }

    def _parse_side_effect(*args):
        return args[0]

    with mock.patch.object(
        results, "_parse_merging_stats", side_effect=_parse_side_effect
    ):
        res = log_cycle_results(results, scaling_script, filter_script)
        # test things have been logged correctly
        cycle_results = res.get_cycle_results()
        assert len(cycle_results) == 1
        assert cycle_results[0]["cumul_percent_removed"] == 100 * 50.0 / 1000.0
        assert cycle_results[0]["n_removed"] == 50
        assert cycle_results[0]["image_ranges_removed"] == [[(6, 10), 0]]
        assert cycle_results[0]["removed_datasets"] == []
        assert cycle_results[0]["delta_cc_half_values"] == [-0.1, 0.1, -0.2, 0.2]
        assert res.get_merging_stats()[0] == "stats_results"
        assert res.initial_n_reflections == 1000

    # add another cycle of results
    with mock.patch.object(
        results, "_parse_merging_stats", side_effect=_parse_side_effect
    ):
        res = log_cycle_results(res, scaling_script, filter_script)
        cycle_results = res.get_cycle_results()
        assert len(cycle_results) == 2
        assert cycle_results[1]["cumul_percent_removed"] == 100 * 2 * 50.0 / 1000.0
        assert cycle_results[1]["n_removed"] == 50
        assert cycle_results[1]["image_ranges_removed"] == [[(6, 10), 0]]
        assert cycle_results[1]["removed_datasets"] == []
        assert cycle_results[0]["delta_cc_half_values"] == [-0.1, 0.1, -0.2, 0.2]
        assert res.get_merging_stats()[1] == "stats_results"
        assert res.initial_n_reflections == 1000


def test_compute_delta_cchalf_returned_results():
    """Test that delta cchalf return necessary values for scale_and_filter."""

    # Check for correct recording of
    # results_summary['per_dataset_delta_cc_half_values']['delta_cc_half_values']
    summary = {}
    delta_cc = {0: -4, 1: 2, 2: -3, 3: -5, 4: 1}
    sorted_data, sorted_ccs = DeltaCCHalf.sort_deltacchalf_values(delta_cc, summary)
    expected_data_order = [3, 0, 2, 4, 1]
    expected_cc_order = [-5, -4, -3, 1, 2]
    assert list(sorted_data) == expected_data_order
    assert list(sorted_ccs) == expected_cc_order
    assert (
        summary["per_dataset_delta_cc_half_values"]["delta_cc_half_values"]
        == expected_cc_order
    )

    # Check for correct recording for dataset mode
    exp = generate_test_experiments(2)
    refls = generate_test_reflections(2)
    ids_to_remove = [0]
    results_summary = {"dataset_removal": {}}
    _ = CCHalfFromDials.remove_datasets_below_cutoff(
        exp, refls, ids_to_remove, results_summary
    )
    assert "experiments_fully_removed" in results_summary["dataset_removal"]
    assert "n_reflections_removed" in results_summary["dataset_removal"]
    assert results_summary["dataset_removal"]["experiments_fully_removed"] == ["0"]
    assert results_summary["dataset_removal"]["n_reflections_removed"] == 10

    # Check for correct recording for image group mode.
    exp = generate_test_experiments(2)
    refls = generate_test_reflections(2)
    ids_to_remove = [0, 1]
    image_group_to_expid_and_range = {
        0: ("0", (1, 5)),
        1: ("0", (6, 10)),
        2: ("1", (1, 5)),
        3: ("1", (6, 10)),
    }
    expids_to_image_groups = {"0": [0, 1], "1": [2, 3]}
    results_summary = {"dataset_removal": {}}
    _ = CCHalfFromDials.remove_image_ranges_below_cutoff(
        exp,
        refls,
        ids_to_remove,
        image_group_to_expid_and_range,
        expids_to_image_groups,
        results_summary,
    )
    assert "experiments_fully_removed" in results_summary["dataset_removal"]
    assert "n_reflections_removed" in results_summary["dataset_removal"]
    assert "image_ranges_removed" in results_summary["dataset_removal"]
    assert results_summary["dataset_removal"]["experiments_fully_removed"] == ["0"]
    assert results_summary["dataset_removal"]["n_reflections_removed"] == 10
    assert [(6, 10), 0] in results_summary["dataset_removal"]["image_ranges_removed"]
    assert [(1, 5), 0] in results_summary["dataset_removal"]["image_ranges_removed"]
    assert len(results_summary["dataset_removal"]["image_ranges_removed"]) == 2


def test_analysis_results_to_from_dict():
    d = {
        "termination_reason": "made up",
        "initial_expids_and_image_ranges": [["foo", [1, 42]], ["bar", [1, 10]]],
        "expids_and_image_ranges": [["foo", [1, 42]]],
        "cycle_results": {"1": {"some stat": -424242}},
        "initial_n_reflections": 424242,
        "final_stats": "some final stats",
    }
    results = AnalysisResults.from_dict(d)
    # The cycle_results dict output by AnalysisResults has integer keys but after
    # conversion to json has str keys. AnalysisResults.from_dict expects str keys,
    # hence do the comparison after converting to/from json
    assert json.loads(json.dumps(results.to_dict())) == d



# LIBTBX_SET_DISPATCHER_NAME dev.dials.generate_tutorial_text

from __future__ import annotations

import argparse
import functools
import json
import os
import pathlib
import shutil
import sys
import tempfile
import time

import procrunner

import dials.util
import dials_data.download


def _command_runner(
    command, output_directory=None, store_command=None, store_output=None, **kwargs
):
    """Run a command and write its output to a defined location"""
    print(" ".join(str(e) for e in command))
    if output_directory and store_command is None:
        store_command = output_directory / f"{command[0]}.cmd"
    if output_directory and store_output is None:
        store_output = output_directory / f"{command[0]}.log"
    if store_command:
        store_command.parent.mkdir(parents=True, exist_ok=True)
        store_command.write_text(" ".join(str(e) for e in command))
    start = time.perf_counter()
    result = procrunner.run(
        command, environment_override={"DIALS_NOBANNER": "1"}, **kwargs
    )
    print(f"running command took {time.perf_counter() - start:.1f} seconds\n")
    assert not result.returncode, "Command execution failed"
    if store_output:
        store_output.parent.mkdir(parents=True, exist_ok=True)
        store_output.write_bytes(result.stdout)


def generate_processing_detail_text_thaumatin(options):
    print("Generating thaumatin processing tutorial output")

    tmpdir = pathlib.Path("tmp-thaumatin")
    tmpdir.mkdir(exist_ok=True)
    outdir = options.output / "thaumatin"
    outdir.mkdir(parents=True, exist_ok=True)
    runcmd = functools.partial(
        _command_runner, output_directory=outdir, working_directory=tmpdir
    )

    df = dials_data.download.DataFetcher()
    runcmd(["dials.import", df("thaumatin_i04", pathlib=True) / "th_8_2_0*cbf"])
    runcmd(["dials.find_spots", "imported.expt", "nproc=4"])
    runcmd(["dials.index", "imported.expt", "strong.refl"])
    runcmd(["dials.refine_bravais_settings", "indexed.expt", "indexed.refl"])
    runcmd(["dials.reindex", "indexed.refl", "change_of_basis_op=a,b,c"])
    runcmd(
        [
            "dials.refine",
            "bravais_setting_9.expt",
            "reindexed.refl",
            "scan_varying=false",
        ]
    )
    runcmd(
        ["dials.refine", "refined.expt", "refined.refl", "scan_varying=true"],
        store_command=outdir / "dials.sv_refine.cmd",
        store_output=outdir / "dials.sv_refine.log",
    )
    runcmd(["dials.integrate", "refined.expt", "refined.refl", "nproc=4"])
    runcmd(["dials.report", "integrated.expt", "integrated.refl"])
    shutil.copy(tmpdir / "dials.report.html", outdir)
    runcmd(["dials.export", "integrated.refl", "integrated.expt"])

    print(f"Updated result files written to {outdir}")
    if not options.keep:
        shutil.rmtree(tmpdir)


def generate_processing_detail_text_mpro_x0692(options):
    print("Generating mpro-x0692 / PDB ID 5REL processing tutorial output")
    # See: https://www.ebi.ac.uk/pdbe/entry/pdb/5REL
    #      https://doi.org/10.5281/zenodo.3730940

    tmpdir = pathlib.Path("tmp-mpro_x0692")
    tmpdir.mkdir(exist_ok=True)
    outdir = options.output / "mpro_x0692"
    outdir.mkdir(parents=True, exist_ok=True)
    runcmd = functools.partial(
        _command_runner, output_directory=outdir, working_directory=tmpdir
    )

    # Find/validate the data input - until we've decided to integrate this
    # into the main release, have a DLS default or otherwise let it be
    # specified via an environment variable.
    df = dials_data.download.DataFetcher()
    runcmd(["dials.import", df("mpro_x0692", pathlib=True) / "Mpro-x0692_1_0*.cbf"])
    runcmd(["dials.find_spots", "imported.expt", "nproc=4"])
    runcmd(["dials.index", "imported.expt", "strong.refl"])
    # Because it's hard to robustly extract the *final* unindexed count
    # using sphinx's built-in literalinclude, we extract it specially here
    extract_last_indexed_spot_count(
        outdir / "dials.index.log", outdir / "dials.index.log.extract_unindexed"
    )
    runcmd(
        [
            "dials.refine_bravais_settings",
            "indexed.expt",
            "indexed.refl",
            "best_monoclinic_beta=False",
        ]
    )
    cb_op = extract_rbs_cb_op(
        tmpdir / "bravais_summary.json",
        outdir / "dials.refine_bravais_settings.log.cb_op",
        2,
    )
    runcmd(["dials.reindex", "indexed.refl", f"change_of_basis_op={cb_op}"])
    runcmd(
        [
            "dials.refine",
            "bravais_setting_2.expt",
            "reindexed.refl",
            "scan_varying=false",
        ]
    )
    runcmd(
        ["dials.refine", "refined.expt", "refined.refl", "scan_varying=true"],
        store_command=outdir / "dials.sv_refine.cmd",
        store_output=outdir / "dials.sv_refine.log",
    )
    runcmd(["dials.integrate", "refined.expt", "refined.refl", "nproc=4"])
    runcmd(
        [
            "dials.symmetry",
            "integrated.expt",
            "integrated.refl",
            "best_monoclinic_beta=False",
        ]
    )
    runcmd(["dials.scale", "symmetrized.expt", "symmetrized.refl"])
    runcmd(["dials.estimate_resolution", "scaled.expt", "scaled.refl"])
    d_min = extract_resolution(outdir / "dials.estimate_resolution.log", "cc_half")
    runcmd(
        ["dials.scale", "scaled.expt", "scaled.refl", f"d_min={d_min:.2f}"],
        store_command=outdir / "dials.scale_cut.cmd",
        store_output=outdir / "dials.scale_cut.log",
    )
    runcmd(["dials.report", "scaled.expt", "scaled.refl"])
    shutil.copy(tmpdir / "dials.report.html", outdir)

    print(f"Updated result files written to {outdir}")
    if not options.keep:
        shutil.rmtree(tmpdir)


def generate_processing_detail_text_betalactamase(options):
    print("Generating Beta-lactamase processing tutorial output")

    tmpdir = pathlib.Path("tmp-betalactamase")
    tmpdir.mkdir(exist_ok=True)
    outdir = options.output / "betalactamase"
    outdir.mkdir(parents=True, exist_ok=True)
    runcmd = functools.partial(
        _command_runner, output_directory=outdir, working_directory=tmpdir
    )

    # Find/validate the data input - until we've decided to integrate this
    # into the main release, have a DLS default or otherwise let it be
    # specified via an environment variable.
    datadir = os.getenv("BETALACTAMASE_TUTORIAL_DATA")
    if datadir:
        datadir = pathlib.Path(datadir)
    else:
        datadir = pathlib.Path(
            os.getenv(
                "CCP4_TUTORIAL_DATA",
                "/dls/i03/data/2017/mx19576-1/tutorial_data/summed/summed/C2sum_1*.cbf.gz",
            )
        ).parent
    if not datadir.is_dir() or not datadir.glob("C2sum_1*.cbf.gz"):
        sys.exit(
            """Error:  Could not find Betalactamase data: skipping text generation.
        Please download C2sum_1 from https://zenodo.org/record/1014387 and extract,
        then set environment variable BETALACTAMASE_TUTORIAL_DATA to the folder with C2sum_1_*.cbf.gz"""
        )
    print(f"Using data: {datadir}")

    runcmd(["dials.import", datadir / "C2sum_1*.cbf.gz"])
    runcmd(["dials.find_spots", "imported.expt", "nproc=4"])
    runcmd(["dials.index", "imported.expt", "strong.refl"])
    # Because it's hard to robustly extract the *final* unindexed count
    # using sphinx's built-in literalinclude, we extract it specially here
    extract_last_indexed_spot_count(
        outdir / "dials.index.log", outdir / "dials.index.log.extract_unindexed"
    )
    runcmd(["dials.refine_bravais_settings", "indexed.expt", "indexed.refl"])
    runcmd(["dials.reindex", "indexed.refl", "change_of_basis_op=a+b,-a+b,c"])
    runcmd(
        [
            "dials.refine",
            "bravais_setting_2.expt",
            "reindexed.refl",
        ]
    )
    runcmd(["dials.integrate", "refined.expt", "refined.refl", "nproc=4"])
    runcmd(["dials.symmetry", "integrated.expt", "integrated.refl"])
    runcmd(["dials.scale", "symmetrized.expt", "symmetrized.refl"])
    runcmd(
        ["dials.scale", "scaled.expt", "scaled.refl", "d_min=1.4"],
        store_command=outdir / "dials.scale_cut.cmd",
        store_output=outdir / "dials.scale_cut.log",
    )
    runcmd(["dials.report", "scaled.expt", "scaled.refl"])
    shutil.copy(tmpdir / "dials.report.html", outdir)

    print(f"Updated result files written to {outdir}")
    if not options.keep:
        shutil.rmtree(tmpdir)


def generate_multi_crystal_symmetry_and_scaling(options):
    print("Generating multi-crystal symmetry analysis and scaling output")

    tmpdir = pathlib.Path(tempfile.mkdtemp("_multi_crystal", dir="."))
    tmpdir.mkdir(exist_ok=True)
    outdir = options.output / "multi_crystal"
    outdir.mkdir(parents=True, exist_ok=True)
    runcmd = functools.partial(
        _command_runner, output_directory=outdir, working_directory=tmpdir
    )

    df = dials_data.download.DataFetcher()
    experiment_files = sorted(
        df("vmxi_proteinase_k_sweeps", pathlib=True).glob("experiments_*.expt")
    )
    reflection_files = sorted(
        df("vmxi_proteinase_k_sweeps", pathlib=True).glob("reflections_*.refl")
    )
    input_files = []
    for src in experiment_files + reflection_files:
        dst = tmpdir / src.name
        os.symlink(src, dst)
        input_files.append(dst.name)
    runcmd(["xia2.multiplex"] + input_files)
    shutil.copy(tmpdir / "xia2.multiplex.html", outdir)
    runcmd(["dials.cosym"] + input_files)
    shutil.copy(tmpdir / "dials.cosym.html", outdir)
    runcmd(["dials.scale", "symmetrized.expt", "symmetrized.refl"])
    runcmd(["dials.estimate_resolution", "scaled.expt", "scaled.refl"])
    d_min = extract_resolution(outdir / "dials.estimate_resolution.log", "cc_half")
    runcmd(
        ["dials.scale", "scaled.expt", "scaled.refl", f"d_min={d_min:.2f}"],
        store_command=outdir / "dials.scale_cut.cmd",
        store_output=outdir / "dials.scale_cut.log",
    )
    runcmd(["dials.compute_delta_cchalf", "scaled.refl", "scaled.expt"])
    runcmd(
        ["dials.scale", "scaled.expt", "scaled.refl", f"d_min={d_min:.2f}"],
        store_command=outdir / "dials.scale_exclude.cmd",
        store_output=outdir / "dials.scale_exclude.log",
    )
    shutil.copy(tmpdir / "dials.scale.html", outdir)
    runcmd(["dials.symmetry", "scaled.expt", "scaled.refl", "laue_group=None"])
    shutil.copy(tmpdir / "dials.symmetry.html", outdir)
    runcmd(["dials.merge", "symmetrized.expt", "symmetrized.refl"])

    print(f"Updated result files written to {outdir}")
    if not options.keep:
        shutil.rmtree(tmpdir)


def find_in_line(string, lines, start=0):
    """Find the next line index containing a given string"""
    for n, line in enumerate(lines[start:], start):
        if string in line:
            return n
    raise RuntimeError(f"Could not find line '{string}' in lines")


def write_extract(destination, start, end, lines):
    """Write lines to a file, in the correct line position, with markers.

    This can be used to provide sphinx with an easily extractable literalinclude
    block, that preserves the correct line numbers from the original file.
    """
    assert start > 0
    out_lines = []
    for n, line in enumerate(lines):
        if n == start - 1:
            out_lines.append("[START_EXTRACT]")
        elif n >= start and n <= end:
            out_lines.append(line.strip())
        elif n == end + 1:
            out_lines.append("[END_EXTRACT]")
        else:
            out_lines.append("")

    destination.write_text("\n".join(out_lines), "latin-1")


def extract_last_indexed_spot_count(source, destination):
    """Extract the last 'unindexed' entry from an index log."""
    lines = source.read_text("latin-1").split("\n")

    # Find the last entry for '# unindexed'
    next_ui = len(lines) - find_in_line("# unindexed", list(reversed(lines))) - 1

    # We now have the last entry for '# unindexed'. Find the end of the table
    end_ui = find_in_line("---", lines, next_ui + 2)

    # Range to extract = (next_ui, end_ui)
    write_extract(destination, next_ui - 1, end_ui, lines)


def extract_resolution(source, method):
    """Extract the resolution from a dials.estimate_resolution log."""
    lines = source.read_text("latin-1").split("\n")

    # Find the Resolution line
    resolution_line = lines[find_in_line(f"Resolution {method}", lines)]
    # Parse and return the suggested resolution value
    return float(resolution_line.split(":")[-1].strip())


def extract_rbs_cb_op(source, destination, solution):
    cb_op = json.loads(source.read_text())[f"{solution}"]["cb_op"]
    write_extract(destination, 1, 3, [cb_op])
    return cb_op


@dials.util.show_mail_handle_errors()
def run(args=None):
    parser = argparse.ArgumentParser(
        description="Generate tutorial logs for DIALS documentation website"
    )
    parser.add_argument("-?", action="help", help=argparse.SUPPRESS)
    parser.add_argument(
        "--beta",
        action="store_true",
        help="Generate betalactamase tutorial logs",
    )
    parser.add_argument(
        "--mpro_x0692",
        action="store_true",
        help="Generate Mpro x0692 tutorial logs",
    )
    parser.add_argument(
        "--thaum",
        action="store_true",
        help="Generate thaumatin tutorial logs",
    )
    parser.add_argument(
        "--multi_crystal",
        action="store_true",
        help="Generate multi-crystal tutorial logs",
    )
    parser.add_argument(
        "--keep",
        action="store_true",
        help="Keep temporary directories on successful generation",
    )
    parser.add_argument(
        "--output",
        action="store",
        type=pathlib.Path,
        default=".",
        help="Write output to this location",
    )
    options = parser.parse_args(args)

    targets = []
    if options.beta:
        targets.append(generate_processing_detail_text_betalactamase)
    if options.mpro_x0692:
        targets.append(generate_processing_detail_text_mpro_x0692)
    if options.thaum:
        targets.append(generate_processing_detail_text_thaumatin)
    if options.multi_crystal:
        targets.append(generate_multi_crystal_symmetry_and_scaling)

    if not targets:
        parser.error(
            "You need to specify at least one dataset\n"
            "to generate logs for (eg. --beta, --thaum)."
        )

    for target in targets:
        target(options)


if __name__ == "__main__":
    run()



from __future__ import annotations

import copy
import logging

import scitbx.matrix
from cctbx import crystal, sgtbx, uctbx
from cctbx.sgtbx.lattice_symmetry import metric_subgroups
from dxtbx.model import Crystal
from scitbx.math import euler_angles_as_matrix

from dials.algorithms.indexing.basis_vector_search import FFT1D, combinations
from dials.algorithms.indexing.symmetry import find_matching_symmetry


def test_combinations(setup_rlp):
    max_cell = 1.3 * max(setup_rlp["crystal_symmetry"].unit_cell().parameters()[:3])
    strategy = FFT1D(max_cell)
    basis_vectors, used = strategy.find_basis_vectors(setup_rlp["rlp"])

    target_symmetry_primitive = (
        setup_rlp["crystal_symmetry"]
        .primitive_setting()
        .customized_copy(space_group_info=sgtbx.space_group().info())
    )
    target_symmetry_sg_only = (
        setup_rlp["crystal_symmetry"]
        .primitive_setting()
        .customized_copy(unit_cell=None)
    )
    target_symmetry_ref = setup_rlp["crystal_symmetry"].as_reference_setting()

    for target_symmetry in (
        setup_rlp["crystal_symmetry"],
        target_symmetry_primitive,
        target_symmetry_sg_only,
        target_symmetry_ref,
    ):

        crystal_models = combinations.candidate_orientation_matrices(
            basis_vectors, max_combinations=50
        )
        crystal_models = list(crystal_models)
        filtered_crystal_models = combinations.filter_known_symmetry(
            crystal_models, target_symmetry=target_symmetry
        )
        filtered_crystal_models = list(filtered_crystal_models)

        assert filtered_crystal_models
        for model in filtered_crystal_models:
            best_subgroup = find_matching_symmetry(
                model.get_unit_cell(), target_symmetry.space_group()
            )
            if target_symmetry.unit_cell() is not None:
                assert best_subgroup["best_subsym"].unit_cell().is_similar_to(
                    setup_rlp["crystal_symmetry"]
                    .as_reference_setting()
                    .best_cell()
                    .unit_cell(),
                    relative_length_tolerance=0.1,
                    absolute_angle_tolerance=5,
                ) or best_subgroup[
                    "best_subsym"
                ].minimum_cell().unit_cell().is_similar_to(
                    setup_rlp["crystal_symmetry"]
                    .as_reference_setting()
                    .best_cell()
                    .minimum_cell()
                    .unit_cell(),
                    relative_length_tolerance=0.1,
                    absolute_angle_tolerance=5,
                )
            else:
                target_sg = (
                    target_symmetry.space_group_info().reference_setting().group()
                )

                subgroups = metric_subgroups(
                    model.get_crystal_symmetry(), max_delta=5, bravais_types_only=False
                )
                assert target_sg.build_derived_patterson_group() in [
                    g["ref_subsym"].space_group() for g in subgroups.result_groups
                ]


def test_filter_known_symmetry_no_matches(caplog):
    caplog.set_level(logging.DEBUG)
    unit_cell = uctbx.unit_cell((10, 10, 10, 90, 90, 90))
    crystal_models = []
    # the reciprocal matrix
    B = scitbx.matrix.sqr(unit_cell.fractionalization_matrix()).transpose()
    for _ in range(10):
        crystal_models.append(
            Crystal(B, space_group=sgtbx.space_group(), reciprocal=True)
        )

    target_symmetry = crystal.symmetry(
        unit_cell=(15, 15, 15, 85, 85, 85), space_group=sgtbx.space_group()
    )
    list(
        combinations.filter_known_symmetry(
            crystal_models, target_symmetry=target_symmetry
        )
    )
    assert "Rejecting crystal model inconsistent with input symmetry" in caplog.text
    assert (
        "No crystal models remaining after comparing with known symmetry" in caplog.text
    )


def test_filter_similar_orientations():
    space_group = sgtbx.space_group()
    unit_cell = space_group.info().any_compatible_unit_cell(volume=1000)

    crystal_models = []
    # the reciprocal matrix
    B = scitbx.matrix.sqr(unit_cell.fractionalization_matrix()).transpose()
    for i in range(10):
        U = euler_angles_as_matrix((i * 10, 0, 0), deg=True)
        crystal_models.append(Crystal(U * B, space_group=space_group, reciprocal=True))

    other_crystal_models = crystal_models[5:]
    crystal_models = crystal_models[:5]

    # there should be no similar orientations to be filtered out
    filtered = list(
        combinations.filter_similar_orientations(crystal_models, other_crystal_models)
    )
    assert filtered == crystal_models

    # now add similar crystal model rotated by 3-degrees
    cryst = copy.deepcopy(other_crystal_models[0])
    cryst.set_U(
        scitbx.matrix.sqr(cryst.get_U()) * euler_angles_as_matrix((3, 0, 0), deg=True)
    )
    crystal_models.append(cryst)
    filtered = combinations.filter_similar_orientations(
        crystal_models, other_crystal_models
    )
    assert list(filtered) == crystal_models[:5]

    # with lower minimum_angular_separation no models should be filtered out
    filtered = combinations.filter_similar_orientations(
        crystal_models, other_crystal_models, minimum_angular_separation=2
    )
    assert list(filtered) == crystal_models



from __future__ import annotations

from dials_algorithms_polygon_clip_ext import *  # noqa: F403; lgtm

__all__ = (  # noqa: F405
    "line_with_rect",
    "quad_with_convex_quad",
    "quad_with_triangle",
    "simple_with_convex",
    "simple_with_rect",
    "triangle_with_convex_quad",
    "triangle_with_triangle",
)



from __future__ import annotations

import json
import logging
import math
import os
import pathlib
import sys
from dataclasses import dataclass, field
from multiprocessing import Pool
from typing import Any, List, Optional, Tuple, Union

from dxtbx.model import Crystal, Experiment, ExperimentList
from libtbx import Auto, phil

from dials.algorithms.indexing import DialsIndexError
from dials.algorithms.indexing.indexer import Indexer
from dials.algorithms.indexing.max_cell import find_max_cell
from dials.array_family import flex
from dials.command_line.combine_experiments import CombineWithReference

RAD2DEG = 180 / math.pi

logger = logging.getLogger("dials")


@dataclass
class InputToIndex:
    reflection_table: flex.reflection_table = None
    experiment: Experiment = None
    parameters: Any = None
    image_identifier: str = ""
    image_no: int = 0
    method_list: List[str] = field(default_factory=list)
    known_crystal_models: Optional[List[Crystal]] = None


@dataclass
class IndexingResult:
    image: str
    image_no: int
    reflection_table: flex.reflection_table = None
    experiments: ExperimentList = None
    n_strong: int = 0
    n_indexed: List[int] = field(default_factory=list)
    identifiers: List[str] = field(default_factory=list)
    rmsd_x: List[float] = field(default_factory=list)
    rmsd_y: List[float] = field(default_factory=list)
    rmsd_dpsi: List[float] = field(default_factory=list)


loggers_to_disable = [
    "dials.algorithms.refinement.reflection_processor",
    "dials.algorithms.refinement.refiner",
    "dials.algorithms.refinement.reflection_manager",
    "dials.algorithms.indexing.stills_indexer",
    "dials.algorithms.indexing.nave_parameters",
    "dials.algorithms.indexing.basis_vector_search.real_space_grid_search",
    "dials.algorithms.indexing.basis_vector_search.combinations",
    "dials.algorithms.indexing.indexer",
]
debug_loggers_to_disable = [
    "dials.algorithms.indexing.symmetry",
    "dials.algorithms.indexing.lattice_search",
    "dials.algorithms.refinement.engine",
]


class manage_loggers(object):

    """
    A contextmanager for reducing logging levels for the underlying code of
    parallel ssx programs.
    If individual_log_verbosity < 2, only timing info will logged.
    If individual_log_verbosity = 2, then the info logs will be shown, but not debug
    logs
    If individual_log_verbosity > 2, then debug logs will be shown.

    """

    def __init__(
        self, individual_log_verbosity, loggers_to_disable, debug_loggers_to_diable=None
    ):
        self.individual_log_verbosity = individual_log_verbosity
        self.loggers = loggers_to_disable
        if debug_loggers_to_disable:
            self.debug_loggers = debug_loggers_to_disable
        else:
            self.debug_loggers = []

    def __enter__(self):
        if self.individual_log_verbosity < 2:
            for name in self.loggers:
                logging.getLogger(name).disabled = True
        elif self.individual_log_verbosity == 2:
            for name in self.loggers:
                logging.getLogger(name).setLevel(logging.INFO)
            for name in self.debug_loggers:
                logging.getLogger(name).setLevel(logging.INFO)

    def __exit__(self, ex_type, ex_value, ex_traceback):
        # Reenable the disabled loggers or reset logging levels.
        if self.individual_log_verbosity < 2:
            for logname in self.loggers:
                logging.getLogger(logname).disabled = False
        elif self.individual_log_verbosity == 2:
            for name in self.loggers:
                logging.getLogger(name).setLevel(logging.DEBUG)
            for name in self.debug_loggers:
                logging.getLogger(name).setLevel(logging.DEBUG)


def index_one(
    experiment: Experiment,
    reflection_table: flex.reflection_table,
    params: phil.scope_extract,
    method_list: List[str],
    image_no: int,
    known_crystal_models: List[Crystal] = None,
) -> Union[Tuple[ExperimentList, flex.reflection_table], Tuple[bool, bool]]:

    elist = ExperimentList([experiment])
    for method in method_list:
        params.indexing.method = method
        idxr = Indexer.from_parameters(
            reflection_table,
            elist,
            params=params,
            known_crystal_models=known_crystal_models,
        )
        try:
            idxr.index()
        except (DialsIndexError, AssertionError) as e:
            logger.info(
                f"Image {image_no+1}: Failed to index with {method} method, error: {e}"
            )
            if method == method_list[-1]:
                return None, None
        else:
            logger.info(
                f"Image {image_no+1}: Indexed {idxr.refined_reflections.size()}/{reflection_table.size()} spots with {method} method."
            )
            return idxr.refined_experiments, idxr.refined_reflections


def wrap_index_one(input_to_index: InputToIndex) -> IndexingResult:
    # First unpack the input and run the function
    expts, table = index_one(
        input_to_index.experiment,
        input_to_index.reflection_table,
        input_to_index.parameters,
        input_to_index.method_list,
        input_to_index.image_no,
        input_to_index.known_crystal_models,
    )

    # Now calculate some useful quantities for the result
    n_strong = input_to_index.reflection_table.get_flags(
        input_to_index.reflection_table.flags.strong
    ).count(True)
    if expts and table:
        result = IndexingResult(
            input_to_index.image_identifier,
            input_to_index.image_no,
            table,
            expts,
            n_strong=n_strong,
        )
        for id_, identifier in table.experiment_identifiers():
            selr = table.select(table["id"] == id_)
            calx, caly, _ = selr["xyzcal.px"].parts()
            obsx, obsy, _ = selr["xyzobs.px.value"].parts()
            delpsi = selr["delpsical.rad"]
            rmsd_x = flex.mean((calx - obsx) ** 2) ** 0.5
            rmsd_y = flex.mean((caly - obsy) ** 2) ** 0.5
            rmsd_z = flex.mean(((delpsi) * RAD2DEG) ** 2) ** 0.5
            n_id_ = calx.size()
            result.n_indexed.append(n_id_)
            result.rmsd_x.append(rmsd_x)
            result.rmsd_y.append(rmsd_y)
            result.rmsd_dpsi.append(rmsd_z)
            result.identifiers.append(identifier)
    else:
        result = IndexingResult(
            input_to_index.image_identifier,
            input_to_index.image_no,
            n_strong=n_strong,
        )

    # If chosen, output a message to json to show live progress
    if input_to_index.parameters.output.nuggets:
        msg = {
            "image_no": input_to_index.image_no + 1,
            "n_indexed": (len(table) if table else 0),
            "n_strong": n_strong,
            "n_cryst": (len(expts) if expts else 0),
            "image": input_to_index.image_identifier,
        }
        with open(
            input_to_index.parameters.output.nuggets
            / f"nugget_index_{msg['image']}.json",
            "w",
        ) as f:
            f.write(json.dumps(msg))

    return result


def index_all_concurrent(
    experiments: ExperimentList,
    reflections: List[flex.reflection_table],
    params: phil.scope_extract,
    method_list: List[str],
) -> Tuple[ExperimentList, flex.reflection_table, dict]:

    input_iterable = []
    results_summary = {
        i: [] for i in range(len(experiments))
    }  # create to give results in order

    # Create a suitable iterable for passing to pool.map
    n = 0
    for iset in experiments.imagesets():
        for i in range(len(iset)):
            refl_index = i + n
            if reflections[refl_index]:
                expt = experiments[refl_index]
                scan = iset.get_scan()
                if scan:
                    idx_0 = i + iset.get_scan().get_batch_offset()  # slicing index
                    new_iset = iset[idx_0 : idx_0 + 1]
                    expt.imageset = new_iset
                input_iterable.append(
                    InputToIndex(
                        reflection_table=reflections[refl_index],
                        experiment=expt,
                        parameters=params,
                        image_identifier=iset.get_image_identifier(i).split("/")[-1],
                        image_no=refl_index,
                        method_list=method_list,
                    )
                )
            else:  # experiments that have already been filtered
                results_summary[refl_index].append(
                    {
                        "Image": iset.get_image_identifier(i).split("/")[-1],
                        "n_indexed": 0,
                        "n_strong": 0,
                    }
                )
        n += len(iset)

    with open(os.devnull, "w") as devnull:
        sys.stdout = devnull  # block printing from rstbx
        with manage_loggers(
            params.individual_log_verbosity,
            loggers_to_disable,
            debug_loggers_to_disable,
        ):
            if params.indexing.nproc > 1:
                with Pool(params.indexing.nproc) as pool:
                    results: List[IndexingResult] = pool.map(
                        wrap_index_one, input_iterable
                    )
            else:
                results: List[IndexingResult] = [
                    wrap_index_one(i) for i in input_iterable
                ]

    sys.stdout = sys.__stdout__
    # prepare tables for output
    indexed_experiments, indexed_reflections = _join_indexing_results(results)

    results_summary = _add_results_to_summary_dict(results_summary, results)

    return indexed_experiments, indexed_reflections, results_summary


def _join_indexing_results(
    results: List[IndexingResult],
) -> Tuple[ExperimentList, flex.reflection_table]:
    indexed_experiments = ExperimentList()
    indexed_reflections = flex.reflection_table()
    n_tot = 0
    for res in results:
        if res.n_indexed:
            indexed_experiments.extend(res.experiments)
            table = res.reflection_table
            ids_map = dict(table.experiment_identifiers())
            for k in table.experiment_identifiers().keys():
                del table.experiment_identifiers()[k]
            table["id"] += n_tot
            for k, v in ids_map.items():
                table.experiment_identifiers()[k + n_tot] = v
            n_tot += len(ids_map.keys())
            indexed_reflections.extend(table)

    indexed_reflections.assert_experiment_identifiers_are_consistent(
        indexed_experiments
    )
    return indexed_experiments, indexed_reflections


def _add_results_to_summary_dict(
    results_summary: dict, results: List[IndexingResult]
) -> dict:
    # convert to results_summary dict current format
    for res in results:
        if res.n_indexed:
            for j in range(len(res.n_indexed)):
                results_summary[res.image_no].append(
                    {
                        "Image": res.image,
                        "identifier": res.identifiers[j],
                        "n_indexed": res.n_indexed[j],
                        "n_strong": res.n_strong,
                        "RMSD_X": res.rmsd_x[j],
                        "RMSD_Y": res.rmsd_y[j],
                        "RMSD_dPsi": res.rmsd_dpsi[j],
                    }
                )
        else:
            results_summary[res.image_no].append(
                {
                    "Image": res.image,
                    "n_indexed": 0,
                    "n_strong": res.n_strong,
                }
            )
    return results_summary


def preprocess(
    experiments: ExperimentList,
    observed: flex.reflection_table,
    params: phil.scope_extract,
) -> Tuple[List[flex.reflection_table], phil.scope_extract, List[str]]:
    reflections = observed.split_by_experiment_id()

    if len(reflections) != len(experiments):
        # spots may not have been found on every image. In this case, the length
        # of the list of reflection tables will be less than the length of experiments.
        # Add in empty items to the list, so that this can be reported on
        obs = set(observed["id"])
        no_refls = set(range(len(experiments))).difference(obs)
        # need to handle both cases where lots have no refls, or only a few do
        full_list = [None] * len(experiments)
        for i, o in enumerate(obs):
            full_list[o] = reflections[i]
        reflections = full_list
        logger.info(f"Filtered {len(no_refls)} images with no spots found.")
        if len(experiments) != len(reflections):
            raise ValueError(
                f"Unequal number of reflection tables {len(reflections)} and experiments {len(experiments)}"
            )

    # Calculate necessary quantities
    n_filtered_out = 0
    filtered_out_images = []
    for i, (refl, experiment) in enumerate(zip(reflections, experiments)):
        if refl:
            if refl.size() >= params.min_spots:
                elist = ExperimentList([experiment])
                refl["imageset_id"] = flex.int(
                    refl.size(), 0
                )  # needed for centroid_px_to_mm
                refl.centroid_px_to_mm(elist)
                refl.map_centroids_to_reciprocal_space(elist)
            else:
                n_filtered_out += 1
                reflections[i] = None
                filtered_out_images.append(i + 1)
        else:
            filtered_out_images.append(i + 1)
    if n_filtered_out:
        logger.info(
            f"Filtered {n_filtered_out} images with fewer than {params.min_spots} spots"
        )
        if params.output.nuggets:
            jstr = json.dumps({"filtered_images": filtered_out_images})
            with open(
                params.output.nuggets / "nugget_index_filtered_images.json", "w"
            ) as f:
                f.write(jstr)

    if n_filtered_out == len(reflections):
        logger.info("All images filtered out, none left to attempt indexing")
        return reflections, params, []

    # Determine the max cell if applicable
    if (params.indexing.max_cell is Auto) and (
        not params.indexing.known_symmetry.unit_cell
    ):
        if params.individual_log_verbosity <= 2:  # suppress the max cell debug log
            logging.getLogger("dials.algorithms.indexing.max_cell").setLevel(
                logging.INFO
            )
        max_cells = []
        for refl in reflections:
            if refl:
                try:
                    max_cells.append(find_max_cell(refl).max_cell)
                except (DialsIndexError, AssertionError):
                    pass
        if not max_cells:
            raise ValueError("Unable to find a max cell for any images")
        logger.info(f"Setting max cell to {max(max_cells):.1f} " + "\u212B")
        params.indexing.max_cell = max(max_cells)

    # Determine which methods to try
    method_list = params.method
    if "real_space_grid_search" in method_list:
        if not params.indexing.known_symmetry.unit_cell:
            logger.info("No unit cell given, real_space_grid_search will not be used")
            method_list.remove("real_space_grid_search")
    methods = ", ".join(method_list)
    pl = "s" if (len(method_list) > 1) else ""
    logger.info(f"Attempting indexing with {methods} method{pl}")

    return reflections, params, method_list


def index(
    experiments: ExperimentList,
    observed: flex.reflection_table,
    params: phil.scope_extract,
) -> Tuple[ExperimentList, flex.reflection_table, dict]:

    if params.output.nuggets:
        params.output.nuggets = pathlib.Path(
            params.output.nuggets
        )  # needed if not going via cli
        if not params.output.nuggets.is_dir():
            logger.warning(
                "output.nuggets not recognised as a valid directory path, no nuggets will be output"
            )
            params.output.nuggets = None

    reflection_tables, params, method_list = preprocess(experiments, observed, params)

    # Do the indexing and generate a summary of results
    indexed_experiments, indexed_reflections, results_summary = index_all_concurrent(
        experiments,
        reflection_tables,
        params,
        method_list,
    )

    # combine beam and detector models if not already
    if (len(indexed_experiments.detectors())) > 1 or (
        len(indexed_experiments.beams())
    ) > 1:
        combine = CombineWithReference(
            detector=indexed_experiments[0].detector, beam=indexed_experiments[0].beam
        )
        elist = ExperimentList()
        for expt in indexed_experiments:
            elist.append(combine(expt))
        indexed_experiments = elist

    return indexed_experiments, indexed_reflections, results_summary



"""
Estimate a resolution limit based on merging statistics calculated in resolution bins.

A number of metrics are supported for estimating a resolution limit,
including:

- `cc_half` (this is the default)
- `isigma` (unmerged <I/sigI>)
- `misigma` (merged <I/sigI>)
- `i_mean_over_sigma_mean` (unmerged <I>/<sigI>)
- `cc_ref` (CC vs provided reference data set)
- `completeness`
- `rmerge`

Resolution estimation is performed by fitting an appropriate curve to the relevant
merging statistics calculated in resolution bins (with a roughly equal number of
reflections per bin). The estimated resolution limit is chosen as the resolution at
which the fitted function equals the specified criteria.

If multiple metrics are requested, the chosen resolution limit will be the lowest
resolution value estimated across the selected metrics.

The fitting functions for the various metrics are defined as follows:

- `cc_half`: fit a tanh function the form (1/2)(1 - tanh(z)) where z = (s - s0)/r, s0 is
  the value of s at the half-falloff value, and r controls the steepness of falloff
- `isigma`, `misigma`, `i_mean_over_sigma_mean`: fit a polynomial to the values
  log(y(x))
- `rmerge`: fit a polynomial to the values log(1/y(x))
- `completeness`: fit a polynomial to the values y(x)

Example use cases
-----------------

Run with defaults on scaled data::

  dials.estimate_resolution scaled.expt scaled.refl

Run with default on scaled unmerged mtz file::

  dials.estimate_resolution scaled_unmerged.mtz

Override the default cc_half cutoff::

  dials.estimate_resolution scaled.expt scaled.refl cc_half=0.1

Use merged <I/sigI> resolution cutoff instead of cc_half::

  dials.estimate_resolution scaled.expt scaled.refl misigma=1.0 cc_half=None

Use unmerged <I/sigI> resolution cutoff in addition to default cc_half::

  dials.estimate_resolution scaled.expt scaled.refl isigma=0.25

Use cc_ref resolution cutoff::

  dials.estimate_resolution cc_ref=0.3 cc_half=None reference=reference.mtz
"""

from __future__ import annotations

import json
import logging
import sys

from jinja2 import ChoiceLoader, Environment, PackageLoader

import libtbx.phil

from dials.util import log, resolution_analysis, show_mail_handle_errors
from dials.util.multi_dataset_handling import parse_multiple_datasets
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files
from dials.util.version import dials_version

logger = logging.getLogger("dials.estimate_resolution")


phil_scope = libtbx.phil.parse(
    """
include scope dials.util.resolution_analysis.phil_defaults

output {
  log = dials.estimate_resolution.log
    .type = path
  html = dials.estimate_resolution.html
    .type = path
  json = None
    .type = path
}
""",
    process_includes=True,
)


@show_mail_handle_errors()
def run(args=None):
    usage = "dials.estimate_resolution [options] (scaled.expt scaled.refl | scaled_unmerged.mtz)"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_reflections=True,
        read_experiments=True,
        check_format=False,
        epilog=__doc__,
    )

    params, options, unhandled = parser.parse_args(
        args=args, return_unhandled=True, show_diff_phil=True
    )

    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )
    if (not reflections or not experiments) and not unhandled:
        parser.print_help()
        return

    if reflections and experiments and unhandled:
        sys.exit(
            "Must provide either scaled unmerged mtz OR dials-format scaled reflections and experiments files"
        )

    # Configure the logging
    log.config(logfile=params.output.log, verbosity=options.verbose)
    logger.info(dials_version())

    if len(unhandled) == 1:
        scaled_unmerged = unhandled[0]
        m = resolution_analysis.Resolutionizer.from_unmerged_mtz(
            scaled_unmerged, params.resolution
        )
    else:
        reflections = parse_multiple_datasets(reflections)
        if len(experiments) != len(reflections):
            sys.exit(
                f"Mismatched number of experiments and reflection tables found: {len(experiments)} & {len(reflections)}."
            )
        m = resolution_analysis.Resolutionizer.from_reflections_and_experiments(
            reflections, experiments, params.resolution
        )

    plots = m.resolution_auto()

    if params.output.html:
        output_html_report(plots, params.output.html)

    if params.output.json:
        with open(params.output.json, "w") as fh:
            json.dump(plots, fh)

    return plots


def output_html_report(plots, filename):
    loader = ChoiceLoader(
        [
            PackageLoader("dials", "templates"),
            PackageLoader("dials", "static", encoding="utf-8"),
        ]
    )
    env = Environment(loader=loader)
    template = env.get_template("simple_report.html")
    html = template.render(
        page_title="dials.estimate_resolution report",
        panel_title="Analysis by resolution",
        panel_id="dials_estimate_resolution",
        graphs=plots,
    )
    with open(filename, "wb") as f:
        f.write(html.encode("utf-8", "xmlcharrefreplace"))


if __name__ == "__main__":
    run()



"""
Phil scope of options for scaling.
"""

from __future__ import annotations

import iotbx.phil

phil_scope = iotbx.phil.parse(
    """
  anomalous = False
    .type = bool
    .help = "Separate anomalous pairs in scaling and error model optimisation."
    .expert_level=0
  overwrite_existing_models = False
    .type = bool
    .help = "If True, create new scaling models for all datasets"
    .expert_level = 0
  reflection_selection {
    method = *quasi_random intensity_ranges use_all random
      .type = choice
      .help = "Method to use when choosing a reflection subset for scaling model"
              "minimisation."
              "The quasi_random option randomly selects reflections groups"
              "within a dataset, and also selects groups which have good"
              "connectedness across datasets for multi-dataset cases. The random"
              "option selects reflection groups randomly for both single"
              "and multi dataset scaling, so for a single dataset"
              "quasi_random == random."
              "The intensity_ranges option uses the E2_range, Isigma_range and"
              "d_range options to the subset of reflections"
              "The use_all option uses all suitable reflections, which may be"
              "slow for large datasets."
    random {
      multi_dataset {
        Isigma_cutoff = 1.0
          .type = float
          .help = "Minimum average I/sigma of reflection groups to use when"
                  "selecting random reflections for minimisation."
      }
      min_groups = 2000
        .type = int
        .help = "The minimum number of symmetry groups to use during"
                "minimisation."
        .expert_level=1
      min_reflections = 50000
        .type = int
        .help = "The minimum number of reflections to use during minimisation."
        .expert_level=1
    }
    best_unit_cell = None
      .type = unit_cell
      .help = "Best unit cell value, to use when performing resolution cutting"
              "and merging statistics. If None, the median cell will be used."
    E2_range = 0.8, 5.0
      .type = floats(size=2)
      .help = "Minimum and maximum normalised E^2 value to used to select a"
              "subset of reflections for minimisation."
      .expert_level = 1
    Isigma_range = -5.0, 0.0
      .type = floats(size=2)
      .help = "Minimum and maximum I/sigma values used to select a subset of"
              "reflections for minimisation. A value of 0.0 for the maximum"
              "indicates that no upper limit should be applied."
      .expert_level = 1
    d_range = None
      .type = floats(size=2)
      .help = "Minimum and maximum d-values used to select a subset of"
              "reflections for minimisation."
      .expert_level = 1
    min_partiality = 0.95
      .type = float
      .help = "Minimum partiality to use when selecting reflections to use"
              "to determine the scaling model and error model."
      .expert_level = 2
    intensity_choice = profile sum *combine
      .alias = intensity
      .type = choice
      .help = "Option to choose from profile fitted or summation intensities, or
               an optimised combination of profile/sum."
      .expert_level = 1
    combine.Imid = None
      .type = floats
      .help = "A list of values to try for the midpoint, for profile/sum combination
               calculation: the value with the lowest Rmeas will be chosen.
               0 and 1 are special values that can be supplied to include profile
               and sum respectively in the comparison."
      .expert_level = 2
    combine.joint_analysis = True
      .type = bool
      .help = "Option of whether to do intensity combination optimisation
              separately (i.e. different Imid per dataset) or joint for
              multiple datasets"
      .expert_level = 2
  }
  weighting {
    weighting_scheme = *invvar
      .type = choice
      .help = "Weighting scheme used during Ih calculation. Weighting schemes
              other than invvar and unity may trigger iterative reweighting
              during minimisation, which may be unstable for certain minimisation
              engines (LBFGS)."
      .expert_level = 2
    error_model {
      include scope dials.algorithms.scaling.error_model.error_model.phil_scope
    }
  }
  cut_data {
    d_min = None
      .type = float
      .help = "Option to apply a high resolution cutoff for the dataset (i.e.
               the chosen reflections have d > d_min)."
      .expert_level = 1
    d_max = None
      .type = float
      .help = "Option to apply a low resolution cutoff for the dataset (i.e.
               the chosen reflections have d < d_max)."
      .expert_level = 1
    partiality_cutoff = 0.4
      .type = float
      .help = "Value below which reflections are removed from the dataset due
               to low partiality."
      .expert_level = 1
    min_isigi = -5
      .type = float
      .help = "Value below which reflections are removed from the dataset due"
              "to low I/sigI in either profile or summation intensity estimates"
      .expert_level = 1
    small_scale_cutoff = 0.001
      .type = float
      .help = "After scaling, remove reflections with scale factors below this"
              "value, in order to avoid reflections with negative scale factors,"
              "and to avoid reflections being greatly scaled when model parameters"
              "refine to very low values."
      .expert_level = 3
  }
  scaling_options {
    check_consistent_indexing = False
      .type = bool
      .help = "If True, run dials.cosym on all data in the data preparation"
              "step, to ensure consistent indexing."
    target_cycle = True
      .type = bool
      .help = "Option to turn of initial round of targeted scaling
               if some datasets are already scaled."
      .expert_level = 2
    only_target = False
      .type = bool
      .help = "Option to only do targeted scaling if some datasets
               are already scaled."
      .expert_level = 2
    only_save_targeted = True
      .type = bool
      .help = "If only_target is true, this option to change whether the dataset
              that is being scaled will be saved on its own, or combined with the
              already scaled dataset."
      .expert_level = 2
    target_model = None
      .type = path
      .help = "Path to cif or pdb file to use to calculate target intensities for"
              "scaling."
              "Deprecated, please use the reference= option instead."
      .expert_level = 2
    target_mtz = None
      .type = path
      .help = "Path to merged mtz file to use as a target for scaling."
              "Deprecated, please use the reference= option instead."
      .expert_level = 2
    reference = None
      .type = path
      .help = "Path to a file to use as a reference for scaling. This can be a"
              "data file containing intensities/amplitudes (.mtz or .cif), or a"
              "model file containing a structure that can be used to calculate"
              "intensities (.cif or .pdb)".
      .expert_level = 2
    nproc = 1
      .type = int(value_min=1)
      .help = "Number of blocks to divide the data into for minimisation.
              This also sets the number of processes to use if the option is
              available."
      .expert_level = 2
    use_free_set = False
      .type = bool
      .help = "Option to use a free set during scaling to check for overbiasing.
              This free set is used to calculate an RMSD, which is shown alongisde
              the 'working' RMSD during refinement, but is not currently used
              to terminate refinement or make any choices on the model."
      .expert_level = 2
    free_set_percentage = 10.0
      .type = float
      .help = "Percentage of symmetry equivalent groups to use for the free set,
              if use_free_set is True."
      .expert_level = 2
    free_set_offset = 0
      .type = int
      .help = "Offset for choosing unique groups for the free set from the whole
               set of unique groups."
      .expert_level = 2
    full_matrix = True
      .type = bool
      .help = "Option to turn off GN/LM refinement round used to determine
               error estimates on scale factors."
      .expert_level = 2
    outlier_rejection = *standard simple
      .type = choice
      .help = "Choice of outlier rejection routine. Standard may take a
        significant amount of time to run for large datasets or high
        multiplicities, whereas simple should be quick for these datasets."
      .expert_level = 1
    outlier_zmax = 6.0
      .type = float(value_min=3.0)
      .help = "Cutoff z-score value for identifying outliers based on their
               normalised deviation within the group of equivalent reflections"
      .expert_level = 1
    emax = 10
      .type = float(value_min = 0)
      .help = "Reject reflections with normalised intensities E^2 > emax^2"
      .expert_level = 2
  }
  """
)



"""
Test the output of targeted scaling - by calling the scale_against_target
scaling_library function and by directly invoking the perform method scaling
of a TargetScaler.
"""

from __future__ import annotations

from math import log

import pytest

from dxtbx.model import Crystal, Experiment, ExperimentList
from libtbx import phil

from dials.algorithms.scaling.model.model import KBScalingModel
from dials.algorithms.scaling.scaler_factory import TargetScalerFactory
from dials.algorithms.scaling.scaling_library import scale_against_target
from dials.array_family import flex
from dials.util.options import ArgumentParser


def generated_target_refl():
    """Test target reflection table."""
    reflections = flex.reflection_table()
    reflections["intensity.prf.value"] = flex.double([1.0, 10.0, 1.0])
    reflections["intensity.prf.variance"] = flex.double([1.0, 10.0, 1.0])
    reflections["intensity.sum.value"] = flex.double([1.0, 10.0, 1.0])
    reflections["intensity.sum.variance"] = flex.double([1.0, 10.0, 1.0])
    reflections["miller_index"] = flex.miller_index(
        [(1, 0, 0), (0, 0, 1), (1, 0, 0)]
    )  # don't change
    reflections["d"] = flex.double([1.0, 2.0, 1.0])
    reflections["partiality"] = flex.double([1.0, 1.0, 1.0])
    reflections.set_flags(flex.bool([True, True, True]), reflections.flags.integrated)
    reflections["id"] = flex.int(3, 0)
    reflections.experiment_identifiers()[0] = str(0)
    return reflections


def generated_refl_to_scale():
    """Generate a reflection table for targeted scaling."""
    reflections = flex.reflection_table()
    reflections["intensity.prf.value"] = flex.double([2.0, 5.0, 2.0, 1.0])
    reflections["intensity.prf.variance"] = flex.double([2.0, 5.0, 2.0, 1.0])
    reflections["intensity.sum.value"] = flex.double([2.0, 5.0, 2.0, 1.0])
    reflections["intensity.sum.variance"] = flex.double([2.0, 5.0, 2.0, 1.0])
    reflections["miller_index"] = flex.miller_index(
        [(1, 0, 0), (0, 0, 1), (1, 0, 0), (10, 0, 0)]
    )  # don't change
    reflections["d"] = flex.double([1.0, 2.0, 1.0, (4.0 / 3.0) ** 0.5])
    reflections["partiality"] = flex.double([1.0, 1.0, 1.0, 1.0])
    reflections.set_flags(
        flex.bool([True, True, True, True]), reflections.flags.integrated
    )
    reflections["id"] = flex.int(4, 1)
    reflections.experiment_identifiers()[1] = str(1)
    return reflections


def test_target_refl():
    """Return the target reflection table."""
    return generated_target_refl()


def test_refl_to_scale():
    """Return the reflection table to scale."""
    return generated_refl_to_scale()


def test_exp(idval=0):
    """Test experiments object."""
    experiments = ExperimentList()
    exp_dict = {
        "__id__": "crystal",
        "real_space_a": [1.0, 0.0, 0.0],
        "real_space_b": [0.0, 1.0, 0.0],
        "real_space_c": [0.0, 0.0, 2.0],
        "space_group_hall_symbol": " C 2y",
    }
    crystal = Crystal.from_dict(exp_dict)
    experiments.append(Experiment(crystal=crystal))
    experiments[0].identifier = str(idval)
    return experiments


@pytest.fixture
def KB_test_param():
    """Generate a params phil scope."""
    phil_scope = phil.parse(
        """
      include scope dials.algorithms.scaling.scaling_options.phil_scope
      include scope dials.algorithms.scaling.model.model.model_phil_scope
      include scope dials.algorithms.scaling.scaling_refiner.scaling_refinery_phil_str
  """,
        process_includes=True,
    )

    parser = ArgumentParser(phil=phil_scope, check_format=False)
    parameters, _ = parser.parse_args(args=[], quick_parse=True, show_diff_phil=False)
    parameters.model = "KB"
    parameters.reflection_selection.method = "use_all"
    return parameters


def test_scale_against_target(KB_test_param):
    """Integration testing of the scale_against_target library function/targeted
    scaling."""
    # Based on input - have Target Ih of 1.0, 10.0 and d of 1.0, 2.0. Then refl to
    # target have I's of 2 and 5, (with the same ds). Therefore for a KB model the
    # problem can be minimised exactly by solving the equations:
    # 2 = K * exp(B/2)
    # 1/2 = K * exp(B/8)
    # Solving these gives the form tested for at the end of this test.
    target_reflections = test_target_refl()
    reflections = test_refl_to_scale()
    target_experiments = test_exp()
    experiments = test_exp(idval=1)
    scaled_reflections = scale_against_target(
        reflections, experiments, target_reflections, target_experiments, KB_test_param
    )
    assert list(scaled_reflections["inverse_scale_factor"]) == pytest.approx(
        [2.0, 0.5, 2.0, 2.0 * (4.0 ** (-1.0 / 3.0))]
    )

    experiments = test_exp()
    experiments.append(test_exp(idval=1)[0])
    experiments[0].scaling_model = KBScalingModel.from_data(KB_test_param, [], [])
    experiments[0].scaling_model.set_scaling_model_as_scaled()
    experiments[1].scaling_model = KBScalingModel.from_data(KB_test_param, [], [])
    target_reflections = test_target_refl()
    reflections = test_refl_to_scale()
    # Repeat the test but calling the TargetScaler directly, to allow inspection
    # of the model components.
    targetscaler = TargetScalerFactory.create(
        KB_test_param, experiments, [target_reflections, reflections]
    )
    targetscaler.perform_scaling()
    assert list(
        targetscaler.unscaled_scalers[0].components["scale"].parameters
    ) == pytest.approx([(4.0 ** (-1.0 / 3.0)) / 2.0])
    assert list(
        targetscaler.unscaled_scalers[0].components["decay"].parameters
    ) == pytest.approx([(log(4.0) * 8.0 / 3.0)])



from __future__ import annotations

from abc import ABC, abstractmethod

import numpy as np
from jinja2 import ChoiceLoader, Environment, PackageLoader

from scitbx.array_family import flex


def generate_html_report(plots_data, filename):
    loader = ChoiceLoader(
        [
            PackageLoader("dials", "templates"),
            PackageLoader("dials", "static", encoding="utf-8"),
        ]
    )
    env = Environment(loader=loader)
    template = env.get_template("simple_report.html")
    html = template.render(
        page_title="DIALS SSX integration report",
        panel_title="Integration plots",
        graphs=plots_data,
    )
    with open(filename, "wb") as f:
        f.write(html.encode("utf-8", "xmlcharrefreplace"))


class SimpleIntegrator(ABC):

    """Define an interface for ssx prediction/integration processing"""

    def __init__(self, params):
        self.params = params
        self.collector = NullCollector()

    @abstractmethod
    def run(self, experiment, table):
        # all gathering/collecting of output data should be optionally done at
        # the run level, so that the calls to individual processing steps are as
        # fast as possible.
        # In general, most output/statistics are calculable on the experiment
        # or reflection table.
        # However the refine step could be expected to return a json with
        # relevant history
        pass

    @abstractmethod
    def preprocess(experiment, reflection_table, *args, **kwargs):
        pass

    @abstractmethod
    def refine(experiment, reflection_table, *args, **kwargs):
        pass

    @abstractmethod
    def predict(experiment, reflection_table, *args, **kwargs):
        pass

    @abstractmethod
    def integrate(experiment, reflection_table, *args, **kwargs):
        pass


class NullCollector(object):

    """
    Defines a null data collector for cases where you don't want
    to record data during the process.
    """

    def __init__(self):
        self.data = {}

    def initial_collect(self, *args, **kwargs):
        pass

    def collect_after_preprocess(self, *args, **kwargs):
        pass

    def collect_after_refinement(self, *args, **kwargs):
        pass

    def collect_after_prediction(self, *args, **kwargs):
        pass

    def collect_after_integration(self, *args, **kwargs):
        pass


class OutputCollector(object):

    """
    Defines a data collector to log common quantities for all algorithm choices
    for an individual image.
    """

    def __init__(self):
        self.data = {}

    # collects general output for reporting, independent of underlying models,
    # for integration of a single image.

    def initial_collect(self, experiment, reflection_table):
        self.data["initial_n_refl"] = reflection_table.size()
        xobs, yobs, _ = reflection_table["xyzobs.px.value"].parts()
        xcal, ycal, _ = reflection_table["xyzcal.px"].parts()
        rmsd = flex.mean((xobs - xcal) ** 2 + (yobs - ycal) ** 2) ** 0.5
        self.data["strong_rmsd"] = rmsd

    def collect_after_preprocess(self, experiment, reflection_table):
        xobs, yobs, _ = reflection_table["xyzobs.px.value"].parts()
        xcal, ycal, _ = reflection_table["xyzcal.px"].parts()
        rmsd = flex.mean((xobs - xcal) ** 2 + (yobs - ycal) ** 2) ** 0.5
        self.data["strong_rmsd_preprocessed"] = rmsd

    def collect_after_prediction(self, predicted, reference):
        matched, _, unmatched = predicted.match_with_reference(reference)
        self.data["n_strong_predicted"] = matched.count(True)
        self.data["n_strong_unpredicted"] = unmatched.size()

    def collect_after_integration(self, experiment, reflection_table):
        sel = reflection_table.get_flags(reflection_table.flags.integrated_sum)
        n_sum = sel.count(True)
        self.data["n_integrated"] = n_sum
        self.data["n_failed"] = reflection_table.size() - n_sum
        I = reflection_table["intensity.sum.value"].select(sel)
        var = reflection_table["intensity.sum.variance"].select(sel)
        if not var.all_gt(0):
            sel2 = var > 0
            I = I.select(sel2)
            var = var.select(sel2)
        self.data["i_over_sigma_overall"] = flex.mean(I / flex.sqrt(var))


class OutputAggregator:

    """
    Simple aggregator class to aggregate data from all images and generate
    json data for output/plotting.
    """

    def __init__(self):
        self.data = {}

    def add_dataset(self, collector, index):
        if collector.data:
            self.data[index] = collector.data

    def make_plots(self):
        # just make some simple plots for now as a test
        if not self.data:
            return {}
        I_over_sigma = [d["i_over_sigma_overall"] for d in self.data.values()]
        n = list(self.data.keys())
        n_integrated = [d["n_integrated"] for d in self.data.values()]
        n_predicted = [d["n_strong_predicted"] for d in self.data.values()]
        n_strong = [
            (d["n_strong_predicted"] + d["n_strong_unpredicted"])
            for d in self.data.values()
        ]
        overall_rmsd = [d["strong_rmsd"] for d in self.data.values()]
        overall_rmsd_preprocessed = [
            d["strong_rmsd_preprocessed"] for d in self.data.values()
        ]

        plots_dict = {
            "I_over_sigma_overall": {
                "data": [
                    (
                        {
                            "x": n,
                            "y": I_over_sigma,
                            "type": "scatter",
                            "mode": "markers",
                        }
                    )
                ],
                "layout": {
                    "title": "Overall I/sigma per image",
                    "xaxis": {"title": "image number"},
                    "yaxis": {"title": "I/sigma"},
                },
            },
            "n_predicted": {
                "data": [
                    {
                        "x": n,
                        "y": n_predicted,
                        "type": "scatter",
                        "mode": "markers",
                        "name": "Number of indexed spots predicted",
                    },
                    {
                        "x": n,
                        "y": n_strong,
                        "type": "scatter",
                        "mode": "markers",
                        "name": "Total number of indexed spots",
                    },
                ],
                "layout": {
                    "title": "Number of predicted reflections per image",
                    "xaxis": {"title": "image number"},
                    "yaxis": {"title": "N. reflections"},
                },
            },
            "n_integrated": {
                "data": [
                    (
                        {
                            "x": n,
                            "y": n_integrated,
                            "type": "scatter",
                            "mode": "markers",
                        }
                    )
                ],
                "layout": {
                    "title": "Number of integrated reflections per image",
                    "xaxis": {"title": "image number"},
                    "yaxis": {"title": "N. reflections"},
                },
            },
            "overall_rmsds": {
                "data": [
                    {
                        "x": n,
                        "y": overall_rmsd,
                        "type": "scatter",
                        "mode": "markers",
                        "name": "Initial 2D rmsd of strong spots",
                    },
                    {
                        "x": n,
                        "y": overall_rmsd_preprocessed,
                        "type": "scatter",
                        "mode": "markers",
                        "name": "2D rmsd after preprocess",
                    },
                ],
                "layout": {
                    "title": "Rmsds of strong (indexed) reflections per image",
                    "xaxis": {"title": "image number"},
                    "yaxis": {"title": "RMSD (px)"},
                },
            },
        }

        value = list(self.data.values())[0]
        mosaicities = {}
        for k in value["profile_model_mosaicity"].keys():
            mosaicities["M_" + k] = np.zeros(shape=(len(self.data),))
        for i, value in enumerate(self.data.values()):
            sigma = value["profile_model_mosaicity"]
            for k, v in sigma.items():
                mosaicities["M_" + k][i] = v
        data = []
        for k, v in mosaicities.items():
            data.append(
                {
                    "x": n,
                    "y": list(v),
                    "type": "scatter",
                    "mode": "markers",
                    "name": k,
                }
            )
        mosaic_plots = {
            "mosaicities": {
                "data": data,
                "layout": {
                    "title": "Profile model mosaicities per image",
                    "xaxis": {"title": "image number"},
                    "yaxis": {"title": "Mosaicity (degrees)"},
                },
            },
        }
        plots_dict.update(mosaic_plots)

        return plots_dict



from __future__ import annotations

import logging

logger = logging.getLogger(__name__)

from libtbx.phil import parse
from scitbx.array_family import flex

from dials.algorithms.refinement import DialsRefineConfigError

phil_str = """
      min_nref_per_parameter = 5
        .help = "the smallest number of reflections per parameter for a"
                "model parameterisation below which the parameterisation will"
                "not be made in full, but the action described below will be"
                "triggered."
        .type = int(value_min=1)

      action = *fail fix remove
        .help = "action to take if there are too few reflections across the"
                "experiments related to a particular model parameterisation."
                "If fail, an exception will be raised and refinement will not"
                "proceed. If fix, refinement will continue but with the"
                "parameters relating to that model remaining fixed at their"
                "initial values. If remove, parameters relating to that model"
                "will be fixed, and in addition all reflections related to"
                "that parameterisation will be removed. This will therefore"
                "remove these reflections from other parameterisations of the"
                "global model too. For example, if a crystal model could not"
                "be parameterised it will be excised completely and not"
                "contribute to the joint refinement of the detector and beam."
                "In the fix mode, reflections emanating from that crystal will"
                "still form residuals and will contribute to detector and beam"
                "refinement."
        .type = choice
"""
phil_scope = parse(phil_str)

# A callback for PredictionParameterisation.get_gradients which will give
# the positions of reflections associated with a particular parameter
def id_associated_refs(result):

    # There are usually 3 parts to results: gradients in X, Y and Z
    vals = list(result.values())
    try:
        vals = [v.as_dense_vector() for v in vals]
    except AttributeError:
        pass

    # Find all non-zero gradients
    vals = [flex.abs(v) > 1e-10 for v in vals]

    # Find any reflection that contributed a non-zero gradient
    val = vals[0]
    for v in vals[1:]:
        val = val | v
    return val


# A callback for PredictionParameterisation.get_gradients which will count
# the number of reflections associated with a particular parameter
def count_associated_refs(result):
    refs = id_associated_refs(result)
    return refs.count(True)


class AutoReduce:
    """Checks for over-parameterisation of models and acts in that case.

    Tests each model parameterisation to ensure there are enough
    reflections in refinement to support that parameterisation. If there are
    not then some action is taken. More details are given in documentation
    within the phil_str alongside this class definition.
    """

    def __init__(
        self,
        options,
        pred_param,
        reflection_manager,
        constraints_manager=None,
        constraints_manager_factory=None,
    ):
        """Initialise the AutoReduce object

        Args:
            options: A PHIL scope containing the auto reduction options
            pred_param: A PredictionParameterisation object containing the
                global model parameterisation
            reflection_manager: A ReflectionManager object tracking the observations
                for use in refinement
            constraints_manager: An optional ConstraintsManager object to allow correct
                determination of the number of reflections per parameter when constraints
                are present
            constraints_manager_factory: An optional ConstraintsManagerFactory, required
                if constraints_manager is present, to recreate the constraints_manager
                if parameters have been fixed
        """

        self._options = options
        self.pred_param = pred_param
        self.param_names = self.pred_param.get_param_names()
        self.reflection_manager = reflection_manager
        self.constraints_manager = constraints_manager
        self.constraints_manager_factory = constraints_manager_factory

    def _nref_per_param(self):
        # Calculate the number of reflections for each parameter, taking
        # constraints into account
        obs = self.reflection_manager.get_obs()
        try:
            self.pred_param.compose(obs)
        except AttributeError:
            pass

        nref_per_param = flex.size_t(
            self.pred_param.get_gradients(obs, callback=count_associated_refs)
        )

        if self.constraints_manager is not None:
            for link in self.constraints_manager.get_constrained_parameter_indices():
                sel = flex.size_t(link)
                total = flex.sum(nref_per_param.select(sel))
                nref_per_param.set_selected(sel, total)

        return nref_per_param

    def check_and_fail(self):
        """Check for too few reflections to support the model parameterisation.

        Test each parameterisation of each type against the reflections it affects.

        Returns:
            None

        Raises:
            DialsRefineConfigError: If there are too few reflections to support
            a parameterisation.
        """

        sel = (
            self._nref_per_param() < self._options.min_nref_per_parameter
        ).iselection()
        if len(sel) > 0:
            names = ", ".join([self.param_names[i] for i in sel])
            msg = f"Too few reflections to parameterise {names}.\n"
            msg += (
                "Try setting "
                "refinement.parameterisation.auto_reduction.action "
                "to fix these parameters (=fix) or additionally remove the "
                "associated reflections (=remove)."
            )
            raise DialsRefineConfigError(msg)

    def check_and_fix(self):
        """Fix parameters when there are too few reflections.

        Test each parameter against the reflections it affects and fix any for
        which there are too few reflections.

        Returns:
            None
        """

        sel = self._nref_per_param() < self._options.min_nref_per_parameter
        isel = sel.iselection()
        if len(isel) > 0:
            names = ", ".join([self.param_names[i] for i in isel])
            msg = f"Too few reflections to parameterise {names}.\n"
            msg += "These parameters will be fixed for refinement."
            logger.warning(msg)
        self.pred_param.fix_params(sel)
        if self.constraints_manager is not None:
            self.constraints_manager = self.constraints_manager_factory()

    def check_and_remove(self):
        """Fix parameters and remove reflections when there are too few reflections.

        Test each parameter against the reflections it affects and fix any for
        which there are too few reflections. In addition, remove all reflections
        that are associated with that parameter to ensure they play no part in
        refinement. This process is iterative.

        Returns:
            None

        Raises:
            DialsRefineConfigError: error if only one single panel detector is present.
        """

        # If there is only one detector in a single experiment, the detector should
        # be multi-panel for remove to make sense
        det_params = self.pred_param.get_detector_parameterisations()
        if len(det_params) == 1:
            n_exp = len(det_params[0].get_experiment_ids())
            if n_exp == 1 and not det_params[0].is_multi_state():
                raise DialsRefineConfigError(
                    "For single experiment, single panel refinement "
                    "auto_reduction.action=remove cannot be used as it could only "
                    "remove all reflections from refinement"
                )

        while True:
            obs = self.reflection_manager.get_obs()
            try:
                self.pred_param.compose(obs)
            except AttributeError:
                pass

            refs_by_parameters = self.pred_param.get_gradients(
                obs, callback=id_associated_refs
            )
            nref_per_param = flex.size_t(
                [refs.count(True) for refs in refs_by_parameters]
            )

            if self.constraints_manager is not None:
                for (
                    link
                ) in self.constraints_manager.get_constrained_parameter_indices():
                    sel = flex.size_t(link)
                    total = flex.sum(nref_per_param.select(sel))
                    nref_per_param.set_selected(sel, total)

            sel = nref_per_param < self._options.min_nref_per_parameter
            if sel.count(True) == 0:
                break

            names = ", ".join([self.param_names[i] for i in sel.iselection()])
            msg = f"Too few reflections to parameterise {names}.\n"
            msg += (
                "These parameters will be fixed for refinement and "
                "the associated reflections will be removed."
            )
            logger.warning(msg)

            self.pred_param.fix_params(sel)

            if self.constraints_manager is not None:
                self.constraints_manager = self.constraints_manager_factory()

            refs_to_filter = flex.bool(len(obs), True)
            for remove, refs in zip(sel, refs_by_parameters):
                if remove:
                    refs_to_filter = refs_to_filter & ~refs

            # only keep refs not associated with this parameterisation
            self.reflection_manager.filter_obs(refs_to_filter)

    def __call__(self):
        """Perform checks and parameter reduction according to the selected option.

        Returns:
            None
        """

        if self._options.action == "fail":
            self.check_and_fail()

        elif self._options.action == "fix":
            self.check_and_fix()

        elif self._options.action == "remove":
            self.check_and_remove()



from __future__ import annotations

from collections import namedtuple

from libtbx.phil import parse
from scitbx import sparse
from scitbx.array_family import flex

from dials.algorithms.refinement import DialsRefineConfigError
from dials.algorithms.refinement.restraints.restraints import (
    LowMemoryMeanUnitCellTie,
    MeanUnitCellTie,
    MedianUnitCellTie,
    SingleUnitCellTie,
)

# PHIL options for unit cell restraints
uc_phil_str = """
restraints
  .help = "Least squares unit cell restraints to use in refinement."
  .expert_level = 1
{
  tie_to_target
    .multiple = True
  {
    values = None
      .type = floats(size=6)
      .help = "Target unit cell parameters for the restraint for this"
              "parameterisation"

    sigmas = None
      .help = "The unit cell target values are associated with sigmas which are"
              "used to determine the weight of each restraint. A sigma of zero"
              "will remove the restraint at that position. If symmetry"
              "constrains two cell dimensions to be equal then only the"
              "smaller of the two sigmas will be kept"
      .type = floats(size=6, value_min=0.)

    id = None
      .help = "Select only the specified experiments when looking up which"
              "parameterisations to apply these restraints to. If an identified"
              "parameterisation affects multiple experiments then the index"
              "of any one of those experiments suffices to restrain that"
              "parameterisation. If None (the default) then the restraints"
              "will be applied to all experiments."
      .type = ints(value_min=0)
  }

  tie_to_group
    .multiple = True
  {
    target = *mean low_memory_mean median
      .type = choice
      .help = "Function to tie group parameter values to"

    sigmas = None
      .help = "The unit cell parameters are associated with sigmas which are"
              "used to determine the weight of each restraint. A sigma of zero"
              "will remove the restraint at that position."
      .type = floats(size=6, value_min=0.)

    id = None
      .help = "Select only the specified experiments when looking up which "
              "parameterisations to apply these restraints to. For every"
              "parameterisation that requires a restraint at least one"
              "experiment index must be supplied. If None (the default) the"
              "restraints will be applied to all experiments."
      .type = ints(value_min=0)
  }
}
"""

uc_phil_scope = parse(uc_phil_str)

# Define a couple of namedtuple types we will use for convenience
ParamIndex = namedtuple("ParamIndex", ["parameterisation", "istart"])
RestraintIndex = namedtuple("RestraintIndex", ["restraint", "istart"])


class RestraintsParameterisation:
    def __init__(
        self,
        detector_parameterisations=None,
        beam_parameterisations=None,
        xl_orientation_parameterisations=None,
        xl_unit_cell_parameterisations=None,
        goniometer_parameterisations=None,
    ):

        if detector_parameterisations is None:
            detector_parameterisations = []
        if beam_parameterisations is None:
            beam_parameterisations = []
        if xl_orientation_parameterisations is None:
            xl_orientation_parameterisations = []
        if xl_unit_cell_parameterisations is None:
            xl_unit_cell_parameterisations = []
        if goniometer_parameterisations is None:
            goniometer_parameterisations = []

        # Keep references to all parameterised models
        self._detector_parameterisations = detector_parameterisations
        self._beam_parameterisations = beam_parameterisations
        self._xl_orientation_parameterisations = xl_orientation_parameterisations
        self._xl_unit_cell_parameterisations = xl_unit_cell_parameterisations
        self._goniometer_parameterisations = goniometer_parameterisations

        # Loop over all parameterisations, extract experiment IDs and record
        # global parameter index for each that tells us which parameters have
        # non-zero derivatives
        self._nparam = 0

        def process(parameterisation_list):
            result = {}
            for parameterisation in parameterisation_list:
                nfree = parameterisation.num_free()
                if nfree == 0:
                    continue
                for iexp in parameterisation.get_experiment_ids():
                    result[iexp] = ParamIndex(parameterisation, self._nparam)
                self._nparam += nfree
            return result

        self._exp_to_det_param = process(self._detector_parameterisations)
        self._exp_to_beam_param = process(self._beam_parameterisations)
        self._exp_to_xlo_param = process(self._xl_orientation_parameterisations)
        self._exp_to_xluc_param = process(self._xl_unit_cell_parameterisations)
        self._exp_to_gon_param = process(self._goniometer_parameterisations)

        # keep a set that will ensure every model parameterisation only gets
        # a single restraint.
        self._param_to_restraint = set()

        # keep lists of restraint objects that we will add
        self._single_model_restraints = []
        self._group_model_restraints = []

    def add_restraints_to_target_xl_unit_cell(self, experiment_id, values, sigma):
        # On input we will have one id value, 6 target values and 6 sigmas.

        # select the right parameterisation, if one exists
        try:
            param_i = self._exp_to_xluc_param[experiment_id]
        except KeyError:
            return

        # fail now if this is already restrained.
        if param_i.parameterisation in self._param_to_restraint:
            raise DialsRefineConfigError(
                "Parameterisation already restrained. Cannot create "
                "additional restraint with experiment {}".format(experiment_id)
            )

        # create new restraint
        tie = SingleUnitCellTie(
            model_parameterisation=param_i.parameterisation, target=values, sigma=sigma
        )

        # add to the restraint list along with the global parameter index
        self._single_model_restraints.append(RestraintIndex(tie, param_i.istart))

        # also add the parameterisation to the set for uniqueness testing
        self._param_to_restraint.add(param_i.parameterisation)

    def add_restraints_to_group_xl_unit_cell(self, target, experiment_ids, sigma):

        # select the right parameterisations, if they exist
        if experiment_ids == "all":
            param_indices = list(self._exp_to_xluc_param.values())
        else:
            param_indices = []
            for exp_id in experiment_ids:
                try:
                    param_indices.append(self._exp_to_xluc_param[exp_id])
                except KeyError:
                    # ignore experiment without a parameterisation
                    pass
        params = [e.parameterisation for e in param_indices]
        istarts = [e.istart for e in param_indices]

        # fail if any of the parameterisations has already been restrained.
        for param in params:
            if param in self._param_to_restraint:
                raise DialsRefineConfigError(
                    "Parameterisation already restrained. Cannot create "
                    "additional group restraint for experiment(s) {}".format(
                        str(experiment_ids)
                    )
                )

        # create new group of restraints
        if target == "mean":
            tie = MeanUnitCellTie(model_parameterisations=params, sigma=sigma)
        elif target == "low_memory_mean":
            tie = LowMemoryMeanUnitCellTie(model_parameterisations=params, sigma=sigma)
        elif target == "median":
            tie = MedianUnitCellTie(model_parameterisations=params, sigma=sigma)
        else:
            raise DialsRefineConfigError(f"target type {target} not available")

        # add to the restraint list along with the global parameter indices
        self._group_model_restraints.append(RestraintIndex(tie, istarts))

    @property
    def num_residuals(self):
        """Get the total number of residuals across all parameterised restraints"""
        n_single = sum(e.restraint.num_residuals for e in self._single_model_restraints)
        n_group = sum(e.restraint.num_residuals for e in self._group_model_restraints)
        return n_single + n_group

    def get_residuals_gradients_and_weights(self):
        residuals = flex.double()
        weights = flex.double()
        row_start = []
        irow = 0

        # process restraints residuals and weights for single models
        for r in self._single_model_restraints:
            res = r.restraint.residuals()
            wgt = r.restraint.weights()
            residuals.extend(flex.double(res))
            weights.extend(flex.double(wgt))
            row_start.append(irow)
            irow += len(res)

        # keep track of the row at the start of group models
        group_model_irow = irow

        # process restraints residuals and weights for groups of models
        for r in self._group_model_restraints:
            residuals.extend(flex.double(r.restraint.residuals()))
            weights.extend(flex.double(r.restraint.weights()))

        # now it is clear how many residuals there are we can set up a sparse
        # matrix for the restraints jacobian
        nrows = len(residuals)
        gradients = sparse.matrix(nrows, self._nparam)

        # assign gradients in blocks for the single model restraints
        for irow, r in zip(row_start, self._single_model_restraints):
            icol = r.istart
            # convert square list-of-lists into a 2D array for block assignment
            grads = flex.double(r.restraint.gradients())
            gradients.assign_block(grads, irow, icol)

        # assign gradients in blocks for the group model restraints
        for r in self._group_model_restraints:
            # loop over the included unit cell models, k
            for k, (icol, grads) in enumerate(zip(r.istart, r.restraint.gradients())):
                irow = group_model_irow
                for grad in grads:
                    gradients.assign_block(grad, irow, icol)
                    irow += grad.n_rows
            group_model_irow = irow

        return residuals, gradients, weights
